import numpy as np
import matplotlib.pyplot as plt
from tabulate import tabulate

from .utils.exceptions import InputError, RockGradingError, user_warning, NotSupportedError
from .core.stability import vandermeer, hudson
from .core.toe import toe_stability
from .core.overtopping import rubble_mound
from .core.scour import scour_protection
from .core.bishop import Bishop
from .core import substructure


class RubbleMound:
    """ General Rubble Mound breakwater class

    Makes a conceptual design for the substructure of a rubble mound
    breakwater. The class computes the necessary nominal diameter and
    rock class of the underlayer, and an filter layer if one is needed.
    Depending on the rock class it is possible that a new variant is
    generated, these are identified by an a, b, c or d. In the attribute
    :py:attr:`variantIDs` a list of generated variants is stored.
    Furthermore, it computes the required crest freeboard of the
    breakwater by using the maximum allowed overtopping discharge from
    the LimitStates with the formulas from EurOtop (2018). Additionally,
    the toe of the breakwater is designed with toe stability formula of
    Van der Meer (1998).

    Parameters
    ----------
    Dn50 : float
        Dn50 of the armour [m]
    Dn50_core : float
        nominal diameter for the stones in the core of the breakwater [m]
    rho : float
        density of the material of the armour layer [kg/m³]
    rho_w : float
        density of water [kg/m³]
    armour_layer : str
        type of armour layer, fully supported armour layers are Rock,
        Xbloc and XblocPlus. Full support means that the rules for the
        underlayer have only been included for these types of armour
        layer. In case another armour layer is used the rule for the
        underlayer, the :py:obj:`filter_rule` must manually be set.
    layers : int
        number of layers in the armour layer
    LimitStates : list
        list of :py:class:`LimitState` ULS, SLS or other defined limit
        states defined with :py:class:`LimitState`. When designing with
        one limit state it must still be entered as a list.
    Grading : :py:class:`RockGrading`
        standard rock grading defined in the NEN-EN 13383-1 or a user
        defined rock grading
    safety : float, optional, default: 1
        safety factor of design (number of standard deviations from the
        mean). Positive values increase the safety, negative values
        decrease the safety of the breakwater
    layers_underlayer : int, optional, default: 2
        number of layers in the underlayer
    slope_toe : tuple, optional, default: (2, 3)
        slope of the toe
    B_toe : float, optional, default: None
        width of the top of the toe in meters. By default the width of
        toe is taken as 3 * Dn50_toe.
    slope : tuple, optional, default: None
        Slope of the armour layer (V, H). For example a slope of 3V:4H
        is defined as (3, 4)
    B : float, optional, default: None
        Crest width [m]
    beta : float, optional, default: 0
        angle between direction of wave approach and a line normal to
        the breakwater (degrees).
    filter_rule : {'Rock', 'Xbloc', 'XblocPlus'}, optional, default: None
        filter rule to use for the substructure of the breakwater, for
        Rock, Xbloc and XblocPlus the correct filter rule is
        automatically selected. In case another type of armour layer is
        used one of these filter rules must be chosen.
    Soil : :py:class:`Soil`, optional, default: None
        by default Soil is None, which means that the geotechnical checks
        are not performed. By specifying a Soil object, the geotechnical
        checks are automatically performed.
    phi : float, optional, default: 40
        internal friction angle of rock [degrees]
    id : int, optional, default: None
        add a unique id to the breakwater

    Attributes
    ----------
    logger : dict
        dict of warnings and messages
    structure : dict
        dictionary with the computed Dn50, rock class, and average Dn50
        of the rock class for each layer and the toe. This dictionary
        includes all variants, use :py:meth:`get_variant` to get the
        parameters of one specific variant. Alternatively,
        :py:meth:`print_variants` can be used to print the details of
        one, multiple or all variants.
    alpha : float
        slope of the structure in radians
    id : int
        unique id of the breakwater
    variantIDs : list
        list with the IDs of the variants generated for this rubble
        mound breakwater.
    Rc : float
        the crest freeboard of the structure [m]
    width_scour : float
        the required length of the scour protection [m]
    bishop : :py:class:`Bishop`
        the :py:class:`Bishop` that was normative in the computation
    F_norm : float
        normative factor of safety computed with :py:class:`Bishop`
    """

    def __init__(
            self, Dn50, Dn50_core, rho, rho_w, armour_layer, layers,
            LimitStates, Grading, safety=1, layers_underlayer=2,
            slope_toe=(2,3), B_toe=None, slope=None, B=None, beta=0,
            filter_rule=None, Soil=None, phi=40, id=None, **kwargs):
        """ See help(RubbleMound) for more info """
        # if not from RockRubbleMound or ConcreteRubbleMound
        if (not isinstance(self, RockRubbleMound)
                and not isinstance(self, ConcreteRubbleMound)):
            # not called from a child class
            # therefore some attributes must be set
            self.logger = {'INFO': [], 'WARNING': []}
            self.structure = {'armour': {'computed Dn50': Dn50,
                                         'class': None,
                                         'class Dn50': Dn50,
                                         'state': None}}
            self.alpha = np.arctan(slope[0]/slope[1])
            self.id = id
            self.variantIDs = ['a', 'b', 'c', 'd']

        # set attribute of bishop and normative F
        self.bishop = None
        self.F_norm = None

        # set input as private attribute
        self._input_arguments = {
            'slope': slope,
            'slope_toe': slope_toe,
            'B': B,
            'armour': armour_layer,
            'Grading': Grading,
            'Dn50_core': Dn50_core
        }

        # check for supported armour layers, and if filter_rule is set
        supported = substructure._supported_armour_layers()
        if armour_layer in supported:
            filter_rule = armour_layer
        elif filter_rule is None:
            supported_rules = ', '.join(supported)
            raise NotSupportedError(
                (f'Filter rule for {armour_layer} is not implemented, set '
                 f'filter rule to use with filter_rule to {supported_rules}'))

        # set the LimitStates as private attribute to use in plot
        self._LimitStates = LimitStates

        # design the first underlayer of the breakwater
        # rho is the density of the material of the armour layer
        computed_dn_u = substructure.underlayer(
            Dn_armour=Dn50, armour_layer=filter_rule, rho=rho,
            rho_rock=Grading.rho)

        # set empty lists to store design values
        class_underlayer, class_dn_u = [], []
        class_filter, class_dn_f, computed_dn_f = [], [], []

        for i, dn in enumerate(computed_dn_u):
            rock_class = Grading.get_class(dn)

            if rock_class in class_underlayer:
                # if rock class is already in the underlayer there is
                # no need to generate an additional variant
                continue
            else:
                # new design and values must thus be saved
                new_underlayer = True

                class_dn = Grading.get_class_dn50(rock_class)
                class_underlayer.append(rock_class)
                class_dn_u.append(class_dn)

                # design a second underlayer/filter layer
                dn_filter_range = substructure.filter_layers(
                    Dn=class_dn, rho=rho)
                for dn_filter in dn_filter_range:
                    if dn_filter > Dn50_core:
                        # computed dn of the filter is larger than the
                        # core, a filter layer is thus needed
                        rock_class = Grading.get_class(dn_filter)
                        if rock_class in class_filter and not new_underlayer:
                            # if rock class is already in the filter layer
                            # there is no need to generate an additional
                            # variant
                            continue
                        else:
                            # new design and values must thus be saved
                            class_dn = Grading.get_class_dn50(rock_class)
                            computed_dn_f.append(dn_filter)
                            class_filter.append(rock_class)
                            class_dn_f.append(class_dn)
                            new_underlayer = False
                    else:
                        # a filter layer is not needed
                        if new_underlayer:
                            # set all values for the filter layer to None
                            computed_dn_f.append(None)
                            class_filter.append(None)
                            class_dn_f.append(None)
                            # set new_underlayer to False so that this
                            # action is not repeated
                            new_underlayer = False
                        else:
                            # still designing for the same underlayer
                            # so no need to set new values
                            continue

        # add underlayer and filter to the structure
        self.structure['underlayer'] = {'computed Dn50': computed_dn_u,
                                        'class': class_underlayer,
                                        'class Dn50': class_dn_u,
                                        'state': 'see armour',
                                        'layers': layers_underlayer}
        if any(class_filter):
            self.structure['filter layer'] = {'computed Dn50': computed_dn_f,
                                              'class': class_filter,
                                              'class Dn50': class_dn_f,
                                              'state': 'see armour',
                                              'layers': layers_underlayer}

        # determine number of variants
        self.variantIDs = self.variantIDs[:len(class_filter)]

        # add generated variants to the logger
        if len(class_underlayer) == 2:
            self.logger['INFO'].append(
                'two rock classes possible for the underlayer, generated new '
                'variant b')
            if len(class_filter) == 4:
                self.logger['INFO'].append(
                    'two rock classes possible for the filter layer, '
                    'generated new variant c')
                self.logger['INFO'].append(
                    'two rock classes possible for the filter layer, '
                    'generated new variant d')
        if len(class_filter) == 3:
            self.logger['INFO'].append(
                'two rock classes possible for the filter layer, generated '
                'new variant c')

        # design toe of the structure
        delta_rock = (Grading.rho - rho_w) / rho_w

        # make a first estimate for the height of the toe
        h_toe = self._estimate_htoe()

        # set variables to store results from normative LimitState
        Dn50_toe, state_toe = 0, 0

        for i, LimitState in enumerate(LimitStates):
            # get values from the LimitState
            Hs = LimitState.get_Hs(definition='H13')
            h = LimitState.h
            Nod = LimitState['Nod']

            # make first estimate for the water level above the toe
            ht_estimate = h - h_toe

            # use while loop since ht depends on dn50 of the toe
            # first set temporary values for the while loop
            Dn50_toe_temp = 0
            compute_toe = True
            counter = 0

            while compute_toe:
                Dn50_toe_computed = toe_stability(
                        Hs=Hs, h=LimitState.h, ht=ht_estimate,
                        Delta=delta_rock, Nod=Nod)

                # check for convergence
                if (abs(Dn50_toe_computed - Dn50_toe_temp) < 0.05
                        or counter > 50):
                    # value has converged, so break loop
                    compute_toe=False

                # replace old value with the new one
                Dn50_toe_temp = Dn50_toe_computed

                # make new estimate for the water level above the toe
                ht_estimate = h - self._estimate_htoe(Dn50=Dn50_toe_computed)

                counter += 1
            # check if computed Dn50 of current LimitState is larger
            # than current normative Dn50
            if Dn50_toe_temp > Dn50_toe:
                # if larger the normative Dn50 must be changed
                Dn50_toe = Dn50_toe_temp
                state_toe = i

        class_toe = Grading.get_class(Dn50_toe)
        class_Dn50 = Grading.get_class_dn50(class_toe)

        self.structure['toe'] = {'computed Dn50': Dn50_toe,
                                 'class': class_toe,
                                 'class Dn50': class_Dn50,
                                 'state': state_toe}

        # determine crest height
        self.Rc, self._state_overtopping = 0, 0

        # check if crest width is more than 3*Dn50 of armour
        if B >= 3*Dn50:
            # no action required
            pass
        else:
            # get armour_layer
            if armour_layer is 'Rock':
                material = 'armourstones'
            else:
                material = 'units'

            user_warning(
                (f'Given crest width is smaller than three {material}, it is '
                  'advised to increase the width to at least '
                 f'{np.round(3*Dn50, 2)} m'))

        # convert beta from deg to rad for overtopping computation
        beta = beta * np.pi/180.

        for i, LimitState in enumerate(LimitStates):
            Hm0 = LimitState.get_Hs(definition='Hm0')
            xi = LimitState.surf_similarity(
                alpha=self.alpha, number='spectral')

            Rc_temp = rubble_mound(
                Hm0=Hm0, q=LimitState['q'], xi_m_min_1=xi, alpha=self.alpha,
                beta=beta, gamma_b=1, gamma_v=1, gam_star=1, Gc=B, Dn50=Dn50,
                armour_layer=armour_layer, layers=layers,
                permeability='permeable', safety=safety)

            # check if computed Rc of current LimitState is larger
            # than current normative Rc
            if Rc_temp > self.Rc:
                # if larger the normative Rc must be changed
                self.Rc = Rc_temp
                self._state_overtopping = i

        # now that the toe is designed the width of the toe can be computed
        # check if a width has been given
        if B_toe is None:
            # set toe width to 3 times Dn50
            self.B_toe = 3 * self.structure['toe']['class Dn50']
        else:
            # user specified width
            self.B_toe = B_toe

            # check if given width is larger than 3 stones
            if B_toe < 3 * self.structure['toe']['class Dn50']:
                user_warning(
                    ('given width of the toe is smaller than three times the '
                     'Dn50 of the toe'))

        # Compute required scour protection
        self.width_scour = 0
        for LimitState in LimitStates:
            w = scour_protection(
                L=LimitState.L(period='Tm'), slope=slope)

            # check if larger than previous value
            if w >= self.width_scour:
                # set w as new width scour
                self.width_scour = w

        # check if a soil has been given
        if Soil is not None:
            # show warning as implementation is not yet verified
            user_warning(
                (f'The implementation of Bishop into {type(self).__name__} '
                  'has not yet been verified with an example'))

            # check if called from ConcreteRubbleMound
            if isinstance(self, ConcreteRubbleMound):
                # raise warning that armour is modelled as rock
                user_warning('The armour layer is modelled as Rock in Bishop')

            # compute the height of the breakwater
            h = LimitStates[self._state_overtopping].h + self.Rc

            # compute horizontal length of the slope
            x = slope[1]*h/slope[0]

            # determine number of slices, with slice width of 1 m
            num_slices = np.round(h, 0)

            n = 0.4

            # compute volumetric weights
            gamma_w = rho_w*9.81/1000
            gamma_r = (1-n)*Grading.rho*9.81/1000
            gamma_r_sat = gamma_r + n*gamma_w

            # set variable to store normative factor of safety
            self.F_norm = 10*10

            # iterate over the LimitStates
            for LimitState in LimitStates:
                # create bishop object
                slip = Bishop(point2=(x,h), wlev=LimitState.h)

                # add soil and rock layer
                slip.add_layer(
                    gamma=Soil.gamma, gamma_sat=Soil.gamma_sat, c=Soil.c,
                    phi=Soil.phi*180/np.pi, name='Subsoil', ymin=-50, ymax=0)
                slip.add_layer(
                    gamma=gamma_r, gamma_sat=gamma_r_sat, c=0, phi=phi,
                    name='Rock', ymin=0, ymax=h)

                # compute factor of safety
                slip.compute(num_slices=int(num_slices), gamma_w=gamma_w)

                # check if normative factor of safety
                if slip.circles[slip.normative].F < self.F_norm:
                    # update normative F and bishop object as attribute
                    self.F_norm = slip.circles[slip.normative].F
                    self.bishop = slip

    def _estimate_htoe(self, Dn50=0):
        """ Method to estimate the height of the toe """
        # set ht variable
        ht = 0

        # get normative layer thickness
        for i, id in enumerate(self.variantIDs):
            # get structure
            structure = self.get_variant(id)

            # get thickness of the layer
            t_armour = self._layer_thickness(
                'armour', self._input_arguments['armour'], structure)
            t_underlayer = self._layer_thickness(
                'underlayer', 'Rock', structure)
            t_filter = self._layer_thickness('filter layer', 'Rock', structure)

            ht_est = t_underlayer + t_filter

            if Dn50 != 0:
                ht_est += np.ceil(t_armour/Dn50) * Dn50

            # check if larger than previous estimate
            if ht_est > ht:
                ht = ht_est

        return ht

    def _validate_variant(self, variants):
        """ Validate the input of the variant

        Parameters
        ----------
        variants : tuple
            variantIDs given as args input
        """
        # check if a variant is specified
        if not variants:
            # no input is given so get the valid args for variant
            valid_args = self.variantIDs
            valid_args.append('all')

            # raise error
            valid = ', '.join(valid_args)
            raise InputError(
                'did not specify which variants to plot, possible arguments '
               f'are {valid}')

        # check if input all is in variants
        if 'all' in variants:
            # set specified variants to all variantIDs
            variants = tuple(self.variantIDs)

        # return the variants
        return variants

    def _layer_thickness(self, layer, material, structure):
        """ Compute the thickness of the layer

        Parameters
        ----------
        layer : {armour, underlayer, filter layer}
            name of the layer for which the layer thickness must be
            computed
        material : str
            material of which the layer is made
        structure : dict
            parameters and values (Dn50, Rock class) for one variant

        Returns
        -------
        float
            thickness of the layer [m]
        """
        # check if layer is in structure
        if layer in structure:
            # get number of layers and Dn50 from the structure
            layers = structure[layer]['layers']
            Dn50 = structure[layer]['class Dn50']

            # get the layer coefficient
            kt = substructure.layer_coefficient(
                material, layers=layers, placement='standard')

            # compute and return the thickness of the layer
            return kt*layers*Dn50

        else:
            # set layer thickness to 0, as layer is not in the structure
            return 0

    def _layers(self, variantID):
        """ compute the coordinates of all layers

        Parameters
        ----------
        variantID : str
            identifier of the variant, see :py:attr:`variantIDs` for a
            list of all generated variants.

        Returns
        -------
        dict
            coordinates of all layers
        """
        # get structure of the current variant
        structure = self.get_variant(variantID)

        # set empty dict to store coordinates in
        coordinates = {}

        # get the slope
        V, H = self._input_arguments['slope']
        V_toe, H_toe = self._input_arguments['slope_toe']

        # compute constant to transformthickness of layer to x and
        # y coordinates, switched V and H because orthogonality
        transform_x = V/np.sqrt(V**2+H**2)
        transform_y = H/np.sqrt(V**2+H**2)

        # get the height and width of the structure
        height = self._LimitStates[self._state_overtopping].h + self.Rc
        B = self._input_arguments['B']

        # determine thickness of the layers
        t_armour = self._layer_thickness(
            'armour', self._input_arguments['armour'], structure)
        t_underlayer = self._layer_thickness('underlayer', 'Rock', structure)
        t_filter = self._layer_thickness('filter layer', 'Rock', structure)

        if t_filter == 0:
            # add scour protection below underlayer on sea side
            t_scour = 2*self._input_arguments['Dn50_core']
        else:
            # filter layer will be used as scour protection
            t_scour = 0

        # compute armour layer
        armour_y1 = t_filter + t_underlayer + t_scour
        armour_y2 = height
        armour_y3 = height

        armour_x1 = -0.5*B - H*(armour_y3-armour_y1)/V
        armour_x2 = -0.5*B
        armour_x3 = 0.5*B
        armour_x4 = 0.5*B + H*armour_y3/V

        # compute line between armour and underlayer
        arm_under_y1 = t_filter + t_underlayer + t_scour
        arm_under_y2 = t_filter + t_underlayer + t_scour
        arm_under_y3 = height - t_armour
        arm_under_y4 = height - t_armour

        arm_under_x5 = (armour_x4 - H*(t_armour * transform_y)/V
                        - t_armour*transform_x)
        arm_under_x4 = arm_under_x5 - H*arm_under_y4/V
        arm_under_x3 = -arm_under_x4
        arm_under_x2 = -H*(arm_under_y3 - arm_under_y2)/V + arm_under_x3

        # compute lower line of the underlayer
        under_y1 = t_filter + t_scour
        under_y2 = t_filter + t_scour
        under_y3 = height - t_armour - t_underlayer
        under_y4 = height - t_armour - t_underlayer

        under_x5 = (arm_under_x5 - H*(t_underlayer * transform_y)/V
                    - t_underlayer*transform_x)
        under_x4 = under_x5 - H*under_y4/V
        under_x3 = -under_x4
        under_x2 = -H*(under_y3 - under_y2)/V + under_x3

        # compute the toe
        Dn50 = structure['toe']['class Dn50']

        # check if armour layer is made out of rock
        # as there is a different toe structure for rock and armour units
        if self._input_arguments['armour'] is 'Rock':
            # compute point where armour layer intersect with the toe
            x = ((abs(arm_under_x2 - armour_x1) * V_toe/H_toe)
                 / (V/H + V_toe/H_toe))
            y = abs(arm_under_x2 - armour_x1)*V*V_toe/(H*V_toe + V*H_toe)

            # determine height of the toe
            htoe = np.ceil(y/Dn50) * Dn50

            toe_low = t_underlayer + t_filter  + t_scour
            toe_top = toe_low + htoe

            toe_x4 = arm_under_x2
            toe_x3 = toe_x4 - H_toe*htoe/V_toe
            toe_x2 = toe_x3 - self.B_toe
            toe_x1 = (toe_x2 - H_toe*htoe/V_toe)

            # change first points of the armour layer
            # so that the armour layer starts at the intersection
            armour_x1 = armour_x1 + x
            armour_y1 = armour_y1 + y

        else:
            # armour units
            htoe = np.ceil(t_armour/Dn50) * Dn50

            toe_low = t_underlayer + t_filter + t_scour
            toe_top = toe_low + htoe

            toe_x4 = armour_x1
            toe_x3 = armour_x1 + H*htoe/V
            toe_x2 = toe_x3 - self.B_toe
            toe_x1 = (toe_x2 - H_toe*htoe/V_toe)

        # check if there is a filter layer
        if t_filter != 0:
            # set points of end of the layers
            arm_under_x1 = toe_x1 - 3*structure['underlayer']['class Dn50']
            arm_under_x0 = arm_under_x1 - H*t_underlayer/V
            arm_under_y0 = t_filter

            # compute the lower line of the filter layer
            filter_y3 = height - t_armour - t_underlayer - t_filter
            filter_y4 = height - t_armour - t_underlayer - t_filter

            filter_x5 = (under_x5 - H*(t_filter * transform_y)/V
                         - t_filter*transform_x)
            filter_x4 = filter_x5 - H*filter_y4/V
            filter_x3 = -filter_x4
            filter_x2 = -H*filter_y3/V + filter_x3
            filter_x1 = arm_under_x0 - self.width_scour
            filter_x0 = filter_x1 - t_filter*H/V

        else:
            # no filter layer
            arm_under_x1 = toe_x1 - 3*structure['underlayer']['class Dn50']
            arm_under_x0 = arm_under_x1 - H*t_underlayer/V
            arm_under_y0 =  t_scour

            # add scour protection
            scour_x2 = arm_under_x0 - self.width_scour
            scour_x1 = scour_x2 - t_scour*H/V

        # add lines to the coordinates
        # are added from left to right and back
        coordinates['armour'] = {
            'x': [armour_x1, armour_x2, armour_x3, armour_x4, arm_under_x5,
                  arm_under_x4, arm_under_x3, arm_under_x2, armour_x1],
            'y': [armour_y1, armour_y2, armour_y3, 0, 0,
                  arm_under_y4, arm_under_y3, arm_under_y2, armour_y1]}

        coordinates['underlayer'] = {
            'x': [arm_under_x0, arm_under_x1, arm_under_x2, arm_under_x3,
                  arm_under_x4, arm_under_x5, under_x5, under_x4, under_x3,
                  under_x2, arm_under_x0],
            'y': [arm_under_y0, arm_under_y1, arm_under_y2, arm_under_y3,
                  arm_under_y4, 0, 0, under_y4, under_y3,
                  under_y2, arm_under_y0]}

        # check if there is a filter to add to coordinates
        if t_filter != 0:
            # add filter to coordinates

            coordinates['filter layer'] = {
                'x': [filter_x0, filter_x1, under_x2, under_x3, under_x4,
                      under_x5, filter_x5, filter_x4, filter_x3, filter_x2,
                      filter_x0],
                'y': [0, t_filter, under_y2, under_y3, under_y4,
                      0, 0, filter_y4, filter_y3, 0, 0]}

            # add core to coordinates
            coordinates['core'] = {
                'x': [filter_x2, filter_x3, filter_x4, filter_x5, filter_x2],
                'y': [0, filter_y3, filter_y4, 0, 0]
            }
        else:
            coordinates['core'] = {
                'x': [scour_x1, scour_x2, under_x2, under_x3, under_x4,
                      under_x5, scour_x1],
                'y': [0, t_scour, t_scour, under_y3, under_y4,
                      0, 0]
            }

        # add the coordinates of the toe
        coordinates['toe'] = {
            'x': [toe_x1, toe_x2, toe_x3, toe_x4, toe_x1],
            'y': [toe_low, toe_top, toe_top, toe_low, toe_low]}

        # return the coordinates of the specified variants
        return coordinates

    def _cost(
            self, *variants, core_price, unit_price, transport_cost,
            output='variant'):
        """ Compute the cost per meter for each variant

        Method to compute the cost of each generated variant, the cost
        is computed per meter

        Parameters
        ----------
        *variants : str
            IDs of the variants to plot, see :py:attr:`variantIDs` for
            a list of all generated variants. If 'all' is in the
            arguments, all variants will be plotted.
        core_price : float
            cost of the core material per m³
        unit_price : float
            the cost of an armour unit per m³
        transport_cost : float
            the cost to transport a m³ of rock from the quarry to the
            project location
        output : {variant, layer, average}
            format of the output dict, variant returns the total cost
            of each variant, layer the cost of each layer for each
            variant and average returns the average cost.

        Returns
        -------
        dict
            the cost

        Raises
        ------
        RockGradingError
            if no pricing is included in the given RockGrading
        """
        # check if transport cost have been given
        if transport_cost is None:
            # no cost thus set cost to zero
            transport_cost = 0

        # validate variants
        variants = self._validate_variant(variants)

        # get the grading, and check if the cost has been added
        Grading = self._input_arguments['Grading']

        if 'price' in Grading[list(Grading.grading.keys())[0]]:
            # pricing has been added
            pass
        else:
            # pricing has not been added, raise error
            raise RockGradingError('There is no pricing in the RockGrading')

        # set empty dict to store the output in
        cost = {}

        # iterate over the generated variants
        for id in variants:
            # get the areas and structure of the variants
            areas = self.area(id)
            structure = self.get_variant(id)

            # iterate over the layers to price each layer
            variant_price = {}
            for layer, area in areas.items():
                if layer is 'core':
                    # core is not included in the structure dict
                    price = ((core_price + transport_cost)
                              * self._input_arguments['Dn50_core'])

                elif (self._input_arguments['armour'] is not 'Rock'
                        and layer is 'armour'):
                    # concrete armour units
                    price = area * unit_price

                else:
                    # layer of the breakwater
                    rock_class = structure[layer]['class']

                    # get the price per meter
                    price = ((Grading[rock_class]['price'] + transport_cost)
                              * area)

                # add to dict
                variant_price[layer] = np.round(price, 2)

            # add to cost dict
            if output is 'variant' or output is 'average':
                # add total cost of all layers
                cost[id] = np.round(np.sum(list(variant_price.values())), 2)
            elif output is 'layer':
                # add the cost of each layer
                cost[id] = variant_price
            else:
                # invalid input
                raise NotSupportedError(
                    (f'Cost can\'t be exported as {output}, must be variant, '
                      'layer or average'))

        # check if average must be computed
        if output is 'average':
            # compute average cost
            cost = {'average': np.round(np.average(list(cost.values())), 2)}

            # check if the cost have only been computed for 1 variant
            if len(variants) == 1:
                # print user_warning and change key into variant
                cost[variants[0]] = cost.pop('average')
                user_warning(
                    ('Computing the average for one variantID, changed key '
                     'average in dict with the specified variantID'))

        return cost

    def get_variant(self, variantID):
        """ Get the dimensions for the specified variant

        Parameters
        ----------
        variantID : str
            identifier of the variant, see :py:attr:`variantIDs` for a
            list of all generated variants.

        Returns
        -------
        dict
            Parameters and values (Dn50, Rock class) for one variant

        Raises
        ------
        KeyError
            If there is no variant with the given identifier
        """
        variant = {'armour': {}, 'underlayer': {}}
        VariantCount = len(self.variantIDs)

        if variantID in self.variantIDs:
            if variantID == 'a':
                key_u = 0
                key_f = 0
            elif variantID == 'b':
                if VariantCount == 4:
                    key_u = 0
                    key_f = 1
                elif 'filter layer' in self.structure:
                    key_f = 1
                    if VariantCount == 2:
                        key_u = 0
                    else:
                        key_u = 1
                elif VariantCount == 2:
                    key_u = 1
            elif variantID == 'c':
                key_u = 1
                if VariantCount >= 3:
                    key_f = 2
            elif variantID == 'd':
                key_u = 1
                key_f = 3
        else:
            raise KeyError(f'Variant with ID = {variantID} is not a variant, '
                           f'generated variants are: {self.variantIDs}')

        # add armour layer
        variant['armour'] = self.structure['armour']

        # add underlayer
        underlayer = self.structure['underlayer']

        for param in underlayer:
            if param == 'state' or param == 'layers':
                variant['underlayer'][param] = underlayer[param]
            else:
                variant['underlayer'][param] = underlayer[param][key_u]

        # add filter layer
        if 'filter layer' in self.structure:
            filter_layer = self.structure['filter layer']
            variant['filter layer'] = {}
            for i, param in enumerate(filter_layer):
                # check because values of these are not list
                if param == 'state' or param == 'layers':
                    # no need to use key_f, because constant for all
                    variant['filter layer'][param] = filter_layer[param]
                else:
                    # param is list and thus use key_f to get correct value
                    variant['filter layer'][param] = filter_layer[param][key_f]

            # check if valid filter layer or all None
            # slice because last two are state and layers
            sliced_values = list(variant['filter layer'].values())[:-2]
            if not any(sliced_values):
                # all values are None, thus delete filter layer
                del variant['filter layer']

        # add toe
        if 'toe' in self.structure:
            variant['toe'] = self.structure['toe']

        return variant

    def plot(self, *variants, wlev=None, save_name=None):
        """ Plot the cross section of the specified breakwater(s)

        Parameters
        ----------
        *variants : str
            IDs of the variants to plot, see :py:attr:`variantIDs` for
            a list of all generated variants. If 'all' is in the
            arguments, all variants will be plotted.
        wlev : str, optional, default: None
            label of the :py:class:`LimitState` from which the water
            level will be plotted. If no value is specified the water
            level from the normative limit state is used, which is the
            normative LimitState from the crest freeboard computation.
        save_name : str, optional, default: None
            if given the cross section is not shown but saved with the
            given name

        Raises
        ------
        InputError
            If no variants are specified or if the label of wlev is not
            a valid label of a :py:class:`LimitState`
        KeyError
            If there is no variant with the given identifier
        """
        # validate variants
        variants = self._validate_variant(variants)

        if wlev is None:
            wlev = self._state_overtopping
        else:
            for i, LimitState in enumerate(self._LimitStates):
                if LimitState.label == wlev:
                    wlev = i
                    break

        # check if wlev has been changed from LimitState label into index
        if isinstance(wlev, str):
            # wlev is still a string so not changed, which means that
            # the specified wlev is not a specified LimitState
            raise InputError('There is no LimitState with the given label')

        V, H = self._input_arguments['slope']

        plt.figure(figsize=(10, 5))

        for i, id in enumerate(variants):
            # set subplot
            if len(variants) == 2:
                plt.subplot(1, 2, i+1)
            elif len(variants) >= 3:
                plt.subplot(2, 2, i+1)

            # get the coordinates
            coordinates = self._layers(id)

            # set xlim_max and xlim_min variable
            xlim_max, xlim_min = 0, 0

            # plot lines
            for layer, lines in coordinates.items():
                plt.plot(lines['x'], lines['y'], color='k')

                # check largest value for xlim
                if np.max(lines['x']) >= xlim_max:
                    # set max as xlim_max
                    xlim_max = np.max(lines['x'])

                # check smallest value for xlim
                if np.min(lines['x']) <= xlim_min:
                    # set min as xlim_min
                    xlim_min = np.min(lines['x'])

            # bottom + wlev
            x_wlev_max = 0.5*self._input_arguments['B'] + H*self.Rc/V

            plt.axhline(y=0, color='k', linewidth=2)
            plt.hlines(
                y=self._LimitStates[wlev].h, xmin=xlim_min*1.2,
                xmax=-x_wlev_max, color='b')
            plt.hlines(
                y=self._LimitStates[wlev].h, xmin=x_wlev_max,
                xmax=xlim_max*1.2, color='b')

            # set xlim and ylim
            ymax = (self._LimitStates[self._state_overtopping].h + self.Rc)*1.2
            plt.xlim(xlim_min*1.2, xlim_max*1.2)
            plt.ylim(-0.5, ymax)

            # add title to the plot
            if save_name is None:
                if isinstance(self.id, int):
                    title = ('Cross section of rubble mound breakwater '
                             f'{self.id}{id}')
                else:
                    title = f'Cross section of rubble mound breakwater {id}'
            else:
                name = save_name.split('/')[-1]
                title = f'Cross section of {name}'

            plt.title(title)

            plt.gca().set_aspect('equal', adjustable='box')
            plt.grid()

        plt.tight_layout()

        if save_name is not None:
            plt.savefig(f'{save_name}.png')
            plt.close()
        else:
            plt.show()

    def print_variant(self, *variants, decimals=3):
        """ Print the details for the specified variant(s)

        This method will print the computed Dn50, rock class, average
        Dn50 of the class, normative LimitState for all layers and
        specified variants.

        Parameters
        ----------
        *variants : str
            IDs of the variants to plot, see :py:attr:`variantIDs` for
            a list of all generated variants. If 'all' is in the
            arguments, all variants will be plotted.
        decimals : int, optional, default: 3
            number of decimal places to round to

        Raises
        ------
        InputError
            If no arguments are specified
        KeyError
            If there is no variant with the given identifier
        """
        # validate variants
        variants = self._validate_variant(variants)

        # loop over the variants
        for id in variants:
            variant = self.get_variant(id)

            # set the name of the table
            if isinstance(self.id, int):
                table_name = f'  Variant {self.id}{id}'
            else:
                table_name = f'  Variant {id}'
            print(table_name)
            table = []

            # iterate over the layers
            for i, (layer, dimensions) in enumerate(variant.items()):
                if i == 0:
                    # set headers of the table
                    headers = list(dimensions.keys())
                    headers.insert(0, 'Layer')
                # add empty list for each row (layer)
                table.append([])
                table[i].append(layer)
                table[i].extend(dimensions.values())
                # print the label of the LimitState instead of the index
                state = dimensions['state']
                index_state = headers.index('state')
                if isinstance(table[i][index_state], int):
                    table[i][index_state] = self._LimitStates[state].label

            print(tabulate(table, headers, tablefmt="github",
                           floatfmt=(f'.{decimals}f')))
            print('')
            print(f'Rc = {np.round(self.Rc, decimals=decimals)} m, designed '
                  f'with {self._LimitStates[self._state_overtopping].label} '
                   'limit state')
            print('\n')

    def print_logger(self, level='warnings'):
        """ Print messages and warnings from the logger

        Parameters
        ----------
        msg_level : {'info', 'warnings'}, optional, default: 'warnings'
            specify print level, highest level is warnings and lowest
            level is info. Note that the info level will also print all
            warnings
        """
        # check if correct input has been given
        if level.lower() not in ['warnings', 'info']:
            raise NotSupportedError(
                f'{level} not implemented, must be info or warnings')

        # print logger
        for type, messages in self.logger.items():
            if level.lower() == 'warnings':
                if type == 'INFO':
                    continue
            print(f'{type}:')
            if messages:
                for message in messages:
                    print(message)
            else:
                print(f'no {type.lower()} messages in log')
            print()

    def area(self, variantID):
        """ Compute the area of all layers

        Method computes the area of each layer using Gauss's area
        formula. Which is given by the following formula:

        .. math::
           \\mathbf{A}=\\frac{1}{2} | \\sum_{i=1}^{n-1} x_{i} y_{i+1}
           +x_{n} y_{1}-\\sum_{i=1}^{n-1} x_{i+1} y_{i}-x_{1} y_{n} |

        Parameters
        ----------
        variantID : str
            identifier of the variant, see :py:attr:`variantIDs` for a
            list of all generated variants.

        Returns
        -------
        dict
            dict with the area of each layer
        """
        # get the coordinates of the layers
        coordinates = self._layers(variantID)

        # iterate over the layers to compute the area
        area = {}
        for layer, coord in coordinates.items():
            # get the x and y coordinates
            x = coord['x']
            y = coord['y']

            # use Gauss's area formula
            A = 0.5*np.abs(np.dot(x,np.roll(y,1))-np.dot(y,np.roll(x,1)))

            # add to area dict
            area[layer] = A

        return area


class RockRubbleMound(RubbleMound):
    """ Design a breakwater with Rock as armour layer

    Makes a conceptual design for a conventional rubble mound breakwater
    with rock as the armour layer, for one or several limit states. The
    following computations are performed:

    - The armour layer is designed with the Van der Meer formulas for
      deep and shallow water (van der Meer, 1988; van Gent et al., 2003).
    - The underlayer is designed by using the rules for the underlayer
    - A filter layer is designed if one is needed, depends on
      :py:obj:`Dn50_core`
    - The toe is designed with the toe stability formula of
      Van der Meer (1998).
    - The crest freeboard is computed with the formula from EurOtop
      (2018)
    - The required width of the scour protection with Sumer and Fredsoe
      (2000)
    - If a :py:class:`Soil` is specified, a slip circle analysis is
      performed with :py:class:`Bishop`

    .. note::
       Depending on the input it might be that more rock classes are
       possible for the underlayer (and filter layer). In case the
       upper bound of the underlayer rule results in a different rock
       class as the lower bound, a new variant is generated. See
       :py:attr:`variantIDs` for a list of generated variants.

    .. note::
       The notional permeability, P, in the van der Meer formula is set
       to a constant value of 0.4. This is due to the fact that the
       substructure of the breakwater will always have an underlayer.

    Parameters
    ----------
    slope : tuple
        Slope of the armour layer (V, H). For example a slope of 3V:4H
        is defined as (3, 4)
    slope_foreshore : tuple
        slope of the foreshore (V, H). For example a slope of 1:100 is
        defined as (1, 100)
    rho_w : float
        density of water [kg/m³]
    B : float
        Crest width [m]
    N : int
        Number of incident waves at the toe of the structure [-]
    LimitState : :py:class:`LimitState` or list of :py:class:`LimitState`
        ULS, SLS or another limit state defined with
        :py:class:`LimitState`
    Grading : :py:class:`RockGrading`
        standard rock grading defined in the NEN-EN 13383-1 or a user
        defined rock grading
    Dn50_core : float
        nominal diameter for the stones in the core of the breakwater [m]
    safety : float, optional, default: 1
        safety factor of design (number of standard deviations from the
        mean)
    slope_toe : tuple, optional, default: (2,3)
        slope of the toe
    B_toe : float, optional, default: None
        width of the top of the toe in meters. By default the width of
        toe is taken as 3 * Dn50_toe.
    beta : float, optional, default: 0
        angle between direction of wave approach and a line normal to
        the breakwater (degrees).
    layers : int, optional, default: 2
        number of layers in the armour layer
    layers_underlayer : int, optional, default: 2
        number of layers in the underlayer
    vdm : {min, max, avg}, optional, default: max
        value to return in case both the deep and shallow water formula
        are valid. min for the lowest value, max for the highest value
        and avg for the average value, default is max.
    Soil : :py:class:`Soil`, optional, default: None
        by default Soil is None, which means that the geotechnical checks
        are not performed. By specifying a Soil object, the geotechnical
        checks are automatically performed.
    phi : float, optional, default: 40
        internal friction angle of rock [degrees]
    id : int, optional, default: None
        add a unique id to the breakwater

    Attributes
    ----------
    logger : dict
        dict of warnings and messages
    structure : dict
        dictionary with the computed Dn50, rock class, and average Dn50
        of the rock class for each layer and the toe. This dictionary
        includes all variants, use :py:meth:`get_variant` to get the
        parameters of one specific variant. Alternatively,
        :py:meth:`print_variant` can be used to print the details of
        one, multiple or all variants.
    alpha : float
        slope of the structure in radians
    id : int
        unique id of the breakwater
    variantIDs : list
        list with the IDs of the variants generated for this rubble
        mound breakwater.
    Rc : float
        the crest freeboard of the structure [m]
    width_scour : float
        the required length of the scour protection [m]
    """
    def __init__(
            self, slope, slope_foreshore, rho_w, B, N, LimitState, Grading,
            Dn50_core, safety=1, slope_toe=(2,3), B_toe=None, beta=0, layers=2,
            layers_underlayer=2, vdm='max', Soil=None, phi=40, id=None,
            **kwargs):
        """ See help(RockRubbleMound) for more info """
        # set logger and structure
        self.logger = {'INFO': [], 'WARNING': []}
        self.structure = {}

        # compute angles
        self.alpha = np.arctan(slope[0]/slope[1])
        slope_foreshore = np.arctan(slope_foreshore[0]/slope_foreshore[1])

        # set id and variantIDs
        self.id = id
        self.variantIDs = ['a', 'b', 'c', 'd']

        # compute relative buoyant density
        delta_rock = (Grading.rho - rho_w)/rho_w

        # set attribute for vandermeer to check validity
        # notional permeability is fixed due to substructure
        self._vandermeer = {
            'N': N, 'P': 0.4, 'Delta': delta_rock}

        # convert single LimitState to list if needed
        if isinstance(LimitState, list):
            LimitStates = LimitState
        else:
            LimitStates = [LimitState]

        # set temporary values to check changes
        Dn50, state = 0, 0

        for i, LimitState in enumerate(LimitStates):
            # design armour layer
            Dn50_temp = vandermeer(
                LimitState=LimitState, Delta=delta_rock, P=0.4, N=N,
                alpha=self.alpha, slope_foreshore=slope_foreshore,
                val=vdm, safety=safety, logger=self.logger)

            # check if computed Dn50 of current LimitState is larger
            # than current normative Dn50
            if Dn50_temp > Dn50:
                # if larger the normative Dn50 must be changed
                Dn50 = Dn50_temp
                state = i

        class_armour = Grading.get_class(Dn50)

        # check dn85/dn15 range with rosin_rammler
        M15 = Grading.rosin_rammler(class_=class_armour, y=0.15)
        dn15 = (M15/Grading.rho)**(1/3)
        M85 = Grading.rosin_rammler(class_=class_armour, y=0.85)
        dn85 = (M85/Grading.rho)**(1/3)
        self._vandermeer['gradation'] = dn85/dn15

        class_Dn50 = Grading.get_class_dn50(class_armour)
        self.structure['armour'] = {'computed Dn50': Dn50,
                                    'class': class_armour,
                                    'class Dn50': class_Dn50,
                                    'state': state,
                                    'layers': layers}

        # design underlayer, filter layer and crest height
        super().__init__(
            Dn50=class_Dn50, Dn50_core=Dn50_core, rho=Grading.rho,
            rho_w=rho_w, armour_layer='Rock', layers=layers,
            LimitStates=LimitStates, Grading=Grading, safety=safety,
            layers_underlayer=layers_underlayer, slope_toe=slope_toe,
            B_toe=B_toe, slope=slope, B=B, beta=beta, id=id, Soil=Soil,
            phi=phi, **kwargs)

    def __str__(self):
        return (f'id.{self.id}: breakwater with rock as armour layer, '
                f'and variants: {self.variantIDs}')

    def cost(
            self, *variants, core_price, transport_cost=None,
            output='variant'):
        """ Compute the cost per meter for each variant

        Method to compute the cost of each generated variant, the cost
        is computed per meter. The cost of the rocks must be specified
        in the RockGrading. If transport cost are not included in the
        price of rocks or core_price it can be given with the argument
        transport_cost.

        Parameters
        ----------
        *variants : str
            IDs of the variants to plot, see :py:attr:`variantIDs` for
            a list of all generated variants. If 'all' is in the
            arguments, all variants will be plotted.
        core_price : float
            cost of the core material per m³
        transport_cost : float, optional, default: None
            the cost to transport a m³ of rock from the quarry to the
            project location
        output : {variant, layer, average}
            format of the output dict, variant returns the total cost
            of each variant, layer the cost of each layer for each
            variant and average returns the average cost.

        Returns
        -------
        dict
            the cost

        Raises
        ------
        RockGradingError
            if no pricing is included in the given RockGrading
        """
        # compute the cost of the concept
        cost = self._cost(
            *variants, core_price=core_price, unit_price=0,
            transport_cost=transport_cost, output=output)

        return cost

    def check_validity(self, decimals=3):
        """ Check if the used parameters are within the validity range

        The Van der Meer equations for deep and shallow water are
        empirical equations, meaning that they are based on experiments.
        Therefore, the Van der Meer equations are, strictly speaking,
        only valid if the parameters are within the range of the
        parameters from the experiments. This method prints a table from
        which the validity range, used value and if the parameter is
        within validity range can be read.

        Parameters
        ----------
        decimals : int, optional, default: 3
            number of decimals
        """
        # set table
        headers = [
            'Parameter    ', 'Validity range', 'Used value', 'in range?']
        table_vdm, table_toe = [], []

        # get the normative LimitState
        state = self.structure['armour']['state']
        state_toe = self.structure['toe']['state']
        LimitState = self._LimitStates[state]

        # check the validity ranges of the van der meer formula
        for info in self.logger['INFO']:
            if LimitState.label in info:
                # msg from the logger with the normative LimitState
                if 'vandermeer_deep' in info:
                    # vandermeer_deep was used
                    print((' Range of validity of parameters in deep water '
                           'formulae by Van der Meer'))

                    # check the slope of the structure
                    slope = self._input_arguments['slope']
                    table_vdm.append(
                        ['tan(alpha)', '1:6 - 1:1.5',
                        f'{slope[0]}:{slope[1]}'])
                    if slope[0]/slope[1] >= 1/6 and slope[0]/slope[1] <= 1/1.5:
                        table_vdm[0].append('yes')
                    else:
                        table_vdm[0].append('NO')

                    # check the number of waves
                    N = self._vandermeer['N']
                    table_vdm.append(['N', '< 7500', np.round(N, decimals)])
                    if N < 7500:
                        table_vdm[1].append('yes')
                    else:
                        table_vdm[1].append('NO')

                    # check wave steepness s_om (based on Tm)
                    s_om = LimitState.s(number='mean')
                    table_vdm.append(
                        ['s_om', '0.01 - 0.06', np.round(s_om, decimals)])
                    if s_om >= 0.01 and s_om <= 0.06:
                        table_vdm[2].append('yes')
                    else:
                        table_vdm[2].append('NO')

                    # surf_similarity parameter based on Tm
                    xi_m = LimitState.surf_similarity(
                        alpha=self.alpha, number='mean')
                    table_vdm.append(
                        ['xi_m', '0.7 - 7', np.round(xi_m, decimals)])
                    if xi_m >= 0.7 and xi_m <= 7:
                        table_vdm[3].append('yes')
                    else:
                        table_vdm[3].append('NO')

                    # delta
                    delta = self._vandermeer['Delta']
                    table_vdm.append(
                        ['Delta', '1 - 2.1', np.round(delta, decimals)])
                    if delta >= 1 and delta <= 2.1:
                        table_vdm[4].append('yes')
                    else:
                        table_vdm[4].append('NO')

                    # relative water depth at the toe
                    Hs = LimitState.get_Hs(
                        definition='H13')
                    h = LimitState.h
                    table_vdm.append(['h/Hs', '> 3', np.round(h/Hs, decimals)])
                    if h/Hs > 3:
                        table_vdm[5].append('yes')
                    else:
                        table_vdm[5].append('NO')

                    # notional permeability
                    P = self._vandermeer['P']
                    table_vdm.append(['P', '0.1 - 0.6', np.round(P, decimals)])
                    if P >= 0.1 and P <= 0.6:
                        table_vdm[6].append('yes')
                    else:
                        table_vdm[6].append('NO')

                    # armourstone gradation
                    gradation = self._vandermeer['gradation']
                    table_vdm.append(
                        ['Dn85/Dn15', '< 2.5', np.round(gradation, decimals)])
                    if gradation < 2.5:
                        table_vdm[7].append('yes')
                    else:
                        table_vdm[7].append('NO')

                    # damage-storm duration ratio
                    Sd = LimitState['Sd']
                    damage_ratio = Sd/np.sqrt(N)
                    table_vdm.append(
                        ['Sd/sqrt(N)', '< 0.9',
                         np.round(damage_ratio, decimals)])
                    if damage_ratio < 0.9:
                        table_vdm[8].append('yes')
                    else:
                        table_vdm[8].append('NO')

                    # stability number
                    Dn50 = self.structure['armour']['computed Dn50']
                    stability = Hs/(delta*Dn50)
                    table_vdm.append(
                        ['Hs/(Delta*Dn50)', '1 - 4',
                         np.round(stability, decimals)])
                    if stability >= 1 and stability <= 4:
                        table_vdm[9].append('yes')
                    else:
                        table_vdm[9].append('NO')

                    # damage level parameter
                    table_vdm.append(['Sd', '1 - 20', np.round(Sd, decimals)])
                    if Sd > 1 and Sd < 30:
                        table_vdm[10].append('yes')
                    else:
                        table_vdm[10].append('NO')

                elif 'vandermeer_shallow' in info:
                    # vandermeer_shallow was used
                    print((' Range of validity of parameters in shallow water '
                           'formulae by Van der Meer'))

                    # check the slope of the structure
                    slope = self._input_arguments['slope']
                    table_vdm.append(
                        ['tan(alpha)', '1:4 - 1:2',
                        f'{slope[0]}:{slope[1]}'])
                    if slope[0]/slope[1] >= 1/4 and slope[0]/slope[1] <= 1/2:
                        table_vdm[0].append('yes')
                    else:
                        table_vdm[0].append('NO')

                    # check the number of waves
                    N = self._vandermeer['N']
                    table_vdm.append(['N', '< 3000', np.round(N, decimals)])
                    if N < 3000:
                        table_vdm[1].append('yes')
                    else:
                        table_vdm[1].append('NO')

                    # check wave steepness s_om (based on Tm)
                    s_om = LimitState.s(number='mean')
                    table_vdm.append(
                        ['s_om', '0.01 - 0.06', np.round(s_om, decimals)])
                    if s_om >= 0.01 and s_om <= 0.06:
                        table_vdm[2].append('yes')
                    else:
                        table_vdm[2].append('NO')

                    # surf_similarity parameter based on Tm
                    xi_m = LimitState.surf_similarity(
                        alpha=self.alpha, number='mean')
                    table_vdm.append(
                        ['xi_m', '1 - 5', np.round(xi_m, decimals)])
                    if xi_m >= 1 and xi_m <= 5:
                        table_vdm[3].append('yes')
                    else:
                        table_vdm[3].append('NO')

                    # surf_similarity parameter based on Tm-1.0
                    xi_m_min_1 = LimitState.surf_similarity(
                        alpha=self.alpha, number='spectral')
                    table_vdm.append(
                        ['xi_m_min_1', '1.3 - 6.5',
                         np.round(xi_m_min_1, decimals)])
                    if xi_m_min_1 >= 1.3 and xi_m_min_1 <= 6.5:
                        table_vdm[4].append('yes')
                    else:
                        table_vdm[4].append('NO')

                    # wave height ratio
                    Hs = LimitState.get_Hs(
                        definition='H13')
                    wave_ratio = LimitState['H2_per']/Hs
                    table_vdm.append(
                        ['H2%/Hs', '1.2 - 1.4', np.round(wave_ratio, decimals)])
                    if wave_ratio >= 1.2 and wave_ratio <= 1.4:
                        table_vdm[5].append('yes')
                    else:
                        table_vdm[5].append('NO')

                    # Deep-water wave height ratio
                    h = LimitState.h
                    table_vdm.append(['Ho/h', '0.25 - 1.5'])
                    try:
                        deep_water_ratio = LimitState['Ho']/h
                        table_vdm[6].append(np.round(deep_water_ratio, decimals))
                        if deep_water_ratio >= 0.25 and deep_water_ratio <= 1.5:
                            table_vdm[6].append('yes')
                        else:
                            table_vdm[6].append('NO')
                    except KeyError:
                        table_vdm[6].append(f'Ho not in {LimitState.label}')

                    # armourstone gradation
                    gradation = self._vandermeer['gradation']
                    table_vdm.append(
                        ['Dn85/Dn15', '1.4 - 2.0',
                         np.round(gradation, decimals)])
                    if gradation >= 1.4 and gradation <= 2:
                        table_vdm[7].append('yes')
                    else:
                        table_vdm[7].append('NO')

                    # core material - armour ratio
                    Dn50 = self.structure['armour']['computed Dn50']
                    Dn50_core = self._input_arguments['Dn50_core']
                    ratio_core = Dn50_core/Dn50
                    table_vdm.append(
                        ['Dn50-core/Dn50', '0 - 0.3',
                         np.round(ratio_core, decimals)])
                    if ratio_core <= 0.3:
                        table_vdm[8].append('yes')
                    else:
                        table_vdm[8].append('NO')

                    # stability number
                    delta = self._vandermeer['Delta']
                    stability = Hs/(delta*Dn50)
                    table_vdm.append(
                        ['Hs/(Delta*Dn50)', '0.5 - 4.5',
                         np.round(stability, decimals)])
                    if stability >= 0.5 and stability <= 4.5:
                        table_vdm[9].append('yes')
                    else:
                        table_vdm[9].append('NO')

                    # damage level parameter
                    Sd = LimitState['Sd']
                    table_vdm.append(['Sd', '< 30', Sd])
                    if Sd < 30:
                        table_vdm[10].append('yes')
                    else:
                        table_vdm[10].append('NO')

        # check validity ranges of toe formula
        # check ratio water depth above toe over water depth
        Dn50_toe = self.structure['toe']['computed Dn50']
        htoe = np.ceil(self._estimate_htoe()/Dn50_toe) * Dn50_toe
        ht = h - htoe

        ratio_depth_toe = ht/h
        table_toe.append(
            ['ht/h', '0.4 - 0.9', np.round(ratio_depth_toe, decimals)])
        if ratio_depth_toe >= 0.4 and ratio_depth_toe <= 0.9:
            table_toe[0].append('yes')
        else:
            table_toe[0].append('NO')

        # ratio ht over Dn50 of the toe
        ratio_dn = ht/Dn50_toe
        table_toe.append(['ht/Dn50', '3 - 25', np.round(ratio_dn, decimals)])
        if ratio_dn >= 3 and ratio_dn <= 25:
            table_toe[1].append('yes')
        else:
            table_toe[1].append('NO')

        # print the table
        print(tabulate(table_vdm, headers, tablefmt="github"))
        print()
        print(' Range of validity of parameters in toe formula')
        print(tabulate(table_toe, headers, tablefmt="github"))


class ConcreteRubbleMound(RubbleMound):
    """ Design a breakwater with concrete armour units as armour layer

    Makes a conceptual design for a conventional rubble mound breakwater
    with armour units as the armour layer, for one or several limit
    states. The following computations are performed:

    - The armour layer is designed with the Hudson formula (Hudson, 1959),
      with :math:`H_{1/3}` as wave height, in accordance with the design
      guidelines for Xbloc and XblocPlus (Delta Marine Consultants, 2018).
    - The underlayer is designed by using the rules for the underlayer
    - A filter layer is designed if one is needed, depends on
      :py:obj:`Dn50_core`
    - The toe is designed with the toe stability formula of
      Van der Meer (1998).
    - The crest freeboard is computed with the formula from EurOtop
      (2018)
    - The required width of the scour protection with Sumer and Fredsoe
      (2000)
    - If a :py:class:`Soil` is specified, a slip circle analysis is
      performed with :py:class:`Bishop`

    .. note::
       Depending on the input it might be that more rock classes are
       possible for the underlayer (and filter layer). In case the
       upper bound of the underlayer rule results in a different rock
       class as the lower bound, a new variant is generated. See
       :py:attr:`variantIDs` for a list of generated variants.

    .. warning::
       Currently only the rules for the underlayer of Rock, Xbloc and
       XblocPlus have been implemented. However, it is possible to
       design with another type of armour units. In this case the filter
       rule must manually be set to Rock, Xbloc or XblocPlus.

    Parameters
    ----------
    slope : tuple
        Slope of the armour layer (V, H). For example a slope of 3V:4H
        is defined as (3, 4)
    slope_foreshore : tuple
        slope of the foreshore (V, H). For example a slope of 1:100 is
        defined as (1, 100)
    B : float
        Crest width [m]
    rho_w : float
        density of water [kg/m³]
    LimitState : :py:class:`LimitState` or list of :py:class:`LimitState`
        ULS, SLS or another limit state defined with
        :py:class:`LimitState`
    ArmourUnit : obj
        armour unit class which inherits from :py:class:`ConcreteArmour`,
        for instance :py:class:`Xbloc` or :py:class:`XblocPlus`
    Grading : :py:class:`RockGrading`
        standard rock grading defined in the NEN-EN 13383-1 or a user
        defined rock grading
    Dn50_core : float
        nominal diameter for the stones in the core of the breakwater [m]
    safety : float, optional, default: 1
        safety factor of design (number of standard deviations from the
        mean)
    slope_toe : tuple, optional, default: (2,3)
        slope of the toe
    B_toe : float, optional, default: None
        width of the top of the toe in meters. By default the width of
        toe is taken as 3 * Dn50_toe.
    beta : float, optional, default: 0
        angle between direction of wave approach and a line normal to
        the breakwater (degrees).
    layers : int, optional, default: 1
        number of layers in the armour layer
    layers_underlayer : int, optional, default: 2
        number of layers in the underlayer
    filter_rule : {'Rock', 'Xbloc', 'XblocPlus'}, optional, default: None
        filter rule to use for the substructure of the breakwater, for
        Rock, Xbloc and XblocPlus the correct filter rule is
        automatically selected. In case another type of armour layer is
        used one of these filter rules must be chosen.
    Soil : :py:class:`Soil`, optional, default: None
        by default Soil is None, which means that the geotechnical checks
        are not performed. By specifying a Soil object, the geotechnical
        checks are automatically performed.
    phi : float, optional, default: 40
        internal friction angle of rock [degrees]
    id : int, optional, default: None
        add a unique id to the breakwater

    Attributes
    ----------
    logger : dict
        dict of warnings and messages
    structure : dict
        dictionary with the computed Dn50, rock class, and average Dn50
        of the rock class for each layer and the toe. This dictionary
        includes all variants, use :py:meth:`get_variant` to get the
        parameters of one specific variant. Alternatively,
        :py:meth:`print_variant` can be used to print the details of
        one, multiple or all variants.
    alpha : float
        slope of the structure in radians
    id : int
        unique id of the breakwater
    variantIDs : list
        list with the IDs of the variants generated for this rubble
        mound breakwater.
    Rc : float
        the crest freeboard of the structure [m]
    width_scour : float
        the required length of the scour protection [m]
    """
    def __init__(
            self, slope, slope_foreshore, B, rho_w, LimitState, ArmourUnit,
            Grading, Dn50_core, safety=1, slope_toe=(2,3), B_toe=None, beta=0,
            layers=1, layers_underlayer=2, filter_rule=None, Soil=None,
            phi=40, id=None, **kwargs):
        """ See help(ConcreteRubbleMound) for more info """
        # set logger and structure
        self.logger = {'INFO': [], 'WARNING': []}
        self.structure = {}

        # compute angles
        self.alpha = np.arctan(slope[0]/slope[1])
        slope_foreshore = np.arctan(slope_foreshore[0]/slope_foreshore[1])

        # compute relative buoyant density
        self.id = id
        self.variantIDs = ['a', 'b', 'c', 'd']

        # compute relative buoyant density
        delta_armour = (ArmourUnit.rho - rho_w)/rho_w

        # convert single LimitState to list if needed
        if isinstance(LimitState, list):
            LimitStates = LimitState
        else:
            LimitStates = [LimitState]

        # determine the type of armour layer
        armour_layer = ArmourUnit.name
        if armour_layer in ['Xbloc', 'XblocPlus']:
            # the formula for Xbloc and XblocPlus is based on the
            # Hudson formula with a fixed slope of 3:4
            angle = np.arctan(3/4)
            filter_rule = armour_layer
        else:
            angle = self.alpha

        # raise error if no filter rule is specified
        if filter_rule is None:
            supported_rules = ', '.join(substructure._supported_armour_layers())
            raise NotSupportedError(
                (f'Filter rule for {armour_layer} is not implemented, set '
                 f'filter rule to use with filter_rule to {supported_rules}'))

        # set temporary values to check changes
        dn, state = 0, 0

        for i, LimitState in enumerate(LimitStates):
            # unpack Hydraulic Conditions
            Hs = LimitState.get_Hs(definition='H13')

            # design armour layer
            dn_temp = hudson(
                H=Hs, Kd=ArmourUnit.kd, Delta=delta_armour, alpha=angle)

            # check if computed Dn50 of current LimitState is larger
            # than current normative Dn50
            if dn_temp > dn:
                # if larger the normative Dn50 must be changed
                dn = dn_temp
                state = i

        V_required = ArmourUnit.get_class(dn)

        self.structure['armour'] = {'computed Dn50': dn,
                                    'class': V_required,
                                    'class Dn50': V_required**(1/3),
                                    'state': state,
                                    'layers': layers}

        # design underlayer, filter layer and crest height
        super().__init__(
            Dn50=V_required**(1/3), Dn50_core=Dn50_core,
            rho=ArmourUnit.rho, rho_w=rho_w, armour_layer=armour_layer,
            layers=layers, LimitStates=LimitStates, Grading=Grading,
            safety=safety, layers_underlayer=layers_underlayer,
            slope_toe=slope_toe, B_toe=B_toe, slope=slope, B=B, beta=beta,
            id=id, filter_rule=filter_rule, Soil=Soil, phi=phi, **kwargs)

        # check if a correction factor must be applied for Xbloc or
        # XblocPlus (DMC design guidelines for Xbloc and XblocPlus)
        correction = getattr(ArmourUnit, 'correction_factor', None)
        if callable(correction):
            # make dict to pass to method correction_factor
            params = {
                'Hs': LimitStates[state].get_Hs('H13'),
                'h': LimitStates[state].h,
                'Rc': self.Rc,
                'occurrence_hs': False,
                'slope': self.alpha,
                'slope_foreshore': slope_foreshore,
                'permeability': 'permeable',
                'Dn': dn,
                'layers': layers,
                'B': B,
                'beta': beta,
            }

            correction_factor = ArmourUnit.correction_factor(
                    logger=self.logger, **params)

        else:
            self.logger['INFO'].append(
                'Given ArmourUnit did not have a method to compute a '
                'correction factor. See documentation how to define a method '
                'to compute a correction factor for a custom armour unit')

            correction_factor = 1.

        # check if a correction factor must be applied
        if correction_factor != 1:
            new_dn = correction_factor**(1/3) * dn

            # replave values in structure with new ones for armour
            new_class = ArmourUnit.get_class(new_dn)
            self.structure['armour'] = {'computed Dn50': new_dn,
                                        'class': new_class,
                                        'class Dn50': new_class**(1/3),
                                        'state': state,
                                        'layers': layers}

            # check if the class of the units has changed
            if new_class != V_required:
                self.logger['INFO'].append(
                    'correction factor resulted in new class, so design will '
                    'be changed')

                # because of the new class the design for the underlayer
                # and filter layer must be changed as well
                super().__init__(
                    Dn50=new_class**(1/3), Dn50_core=Dn50_core,
                    rho=ArmourUnit.rho, rho_w=rho_w, armour_layer=armour_layer,
                    layers=layers, LimitStates=LimitStates, Grading=Grading,
                    safety=safety, layers_underlayer=layers_underlayer,
                    slope_toe=slope_toe, B_toe=B_toe, slope=slope, B=B,
                    beta=beta, id=id, filter_rule=filter_rule, Soil=Soil,
                    phi=phi, **kwargs)

            else:
                # same class so design does not have to be changed
                self.logger['INFO'].append(
                    'correction factor did not result in a new class, so '
                    'design will not be changed')

    def __str__(self):
        return (f'id.{self.id}: breakwater with armour units as armour layer, '
                f'and variants: {self.variantIDs}')

    def cost(
            self, *variants, core_price, unit_price, transport_cost=None,
            output='variant'):
        """ Compute the cost per meter for each variant

        Method to compute the cost of each generated variant, the cost
        is computed per meter. The cost of the rocks in the substructure
        must be specified in the RockGrading. If transport cost are not
        included in the price of rocks or core_price it can be given
        with the argument transport_cost.

        .. note::
           The transport_cost are not added to the price of the armour
           layer. The assumption has been made that the cost of
           producing and transporting the armour units is included in
           the unit_price.

        Parameters
        ----------
        *variants : str
            IDs of the variants to plot, see :py:attr:`variantIDs` for
            a list of all generated variants. If 'all' is in the
            arguments, all variants will be plotted.
        core_price : float
            cost of the core material per m³
        unit_price : float
            the cost of an armour unit per m³
        transport_cost : float, optional, default: None
            the cost to transport a m³ of rock from the quarry to the
            project location
        output : {variant, layer, average}
            format of the output dict, variant returns the total cost
            of each variant, layer the cost of each layer for each
            variant and average returns the average cost.

        Returns
        -------
        dict
            the cost

        Raises
        ------
        RockGradingError
            if no pricing is included in the given RockGrading
        """
        # compute the cost of the concept
        cost = self._cost(
            *variants, core_price=core_price, unit_price=unit_price,
            transport_cost=transport_cost, output=output)

        return cost



# import all functions and classes from the core required to design a bw
from .battjes import BattjesGroenendijk
from .bishop import Bishop, SlipCircle
from .goda import goda_wave_heights, Goda
from .overtopping import gamma_f, gamma_beta, rubble_mound, vertical_deep
from .overtopping import vertical_no_breaking, vertical_normal, vertical_low
from .overtopping import composite_normal, composite_low, vertical
from .scour import scour_protection
from .soil import Soil
from .stability import vandermeer_deep, vandermeer_shallow, vandermeer, hudson
from .substructure import underlayer, filter_layers, layer_coefficient
from .toe import toe_berm_stability, toe_stability

# import functions and classes to compute wave heights
from ..utils.wave import shoaling_coefficient, dispersion
from ..conditions import LimitState

# import materials
from ..material import RockGrading, Xbloc, XblocPlus, ConcreteArmour



import context
import unittest
import warnings
import numpy as np

from breakwater.core.goda import goda_wave_heights
from breakwater.conditions import LimitState
from breakwater.material import RockGrading, Xbloc
from breakwater.caisson import Caisson


class TestCaisson(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        warnings.filterwarnings('ignore')

        H13, Hmax = goda_wave_heights(
            h=15, d=11, Ho=5.7, T=7.28, slope_foreshore=(1,100))
        H13_SLS, Hmax_SLS = goda_wave_heights(
            h=13, d=8, Ho=4, T=7.18, slope_foreshore=(1,100))

        ULS = LimitState(
            deep_water=False, h=15, Nod=0.5, Sd=2, Ho=5, q=10, Hm0=5.22,
            H13=H13, Hmax=Hmax, T_m_min_1=8.66, Tp=7.87, Tm=6.3, T13=7.28,
            label='ULS')
        SLS = LimitState(
            deep_water=False, h=13, Nod=0.5, Sd=2, Ho=4, q=20, Hm0=4.22,
            H13=H13_SLS, Hmax=Hmax_SLS, T_m_min_1=8.66, Tp=7.87, Tm=6.3,
            T13=7.18, label='SLS')

        NEN = RockGrading()

        cls.bw = Caisson(
            Pc=0.6, rho_c=2400, rho_fill=1600, rho_w=1000, Bm=8, hb=4, mu=0.5,
            layers=2, BermMaterial=NEN, LimitState=[SLS, ULS], safety=1,
            slope_foreshore=(1,100), SF_sliding=1.2, SF_turning=1.2,
            beta=0, slope_foundation=(2,3))

    def test_structure_caisson(self):
        caisson_structure = self.bw.structure['caisson']
        computed = {
            'hb': 4, 'h_acc': 11, 'Pc': 0.6, 'd': 8.333,
            'Rc': 14.222, 'state_overtop': 1,
            'B': 18.888, 'state_goda': 1, 'Bm': 8}
        for i, (param, val) in enumerate(caisson_structure.items()):
            self.assertAlmostEqual(val, list(computed.values())[i], 3)

    def test_structure_armour(self):
        armour = self.bw.structure['armour']
        computed = [1.334, 'HMA_6000/10000', 1.445, 1, 2]
        for i, (param, val) in enumerate(armour.items()):
            self.assertAlmostEqual(val, computed[i], 3)

    def test_structure_underlayer(self):
        underlayer = self.bw.structure['foundation']
        computed = [[0.586, 0.671],
                    ['HMA_300/1000', 'HMA_1000/3000'],
                    [0.615, 0.895],
                    'see armour']
        for i, (param, vals) in enumerate(underlayer.items()):
            for j, val in enumerate(vals):
                self.assertAlmostEqual(val, computed[i][j], 3)

    def test_effective_width_foundation(self):
        B_eff = self.bw._effective_width_foundation(
            rho_w=1000, hb=4, B=18.895, m=991289)
        self.assertAlmostEqual(B_eff, 22.309, 3)


if __name__ == '__main__':
    unittest.main()



import numpy as np

from ..utils.exceptions import NotSupportedError

def xi_critical(Cpl, Cs, P, alpha):
    """ Critical value of the surf-similarity parameter for Van der Meer

    The critical value of the surf-similarity parameter is used in the
    Van der Meer formulas to determine the transition from plunging to
    surging waves.

    Parameters
    ----------
    P : float
        Notional permeability of the structure [-]
    Cpl : float
        Model constant for plunging waves [-]
    Cs : float
        Model constant for surging waves [-]
    alpha : float
        Slope of the structure [rad]

    Returns
    -------
    float
        The critical value of the surf-similarity parameter [-]
    """
    xi_cr = (Cpl/Cs * P**0.31 * np.sqrt(np.tan(alpha)))**(1/(P+0.5))
    return xi_cr


def vandermeer_deep(Hs, Delta, P, Sd, N, xi_m, alpha, safety=1):
    """ Van der Meer formulae for deep water conditions

    For deep water conditions Van der Meer (1988) derived formulae to
    predict the stability of armourstone on uniform armourstone slopes
    with crests above the maximum run-up level. These formulae are only
    valid in deep water conditions, which is defined as
    :math:`h > 3 \\cdot H_{s-toe}`, where :math:`h` is the water depth
    on the toe of the structure and :math:`H_{s-toe}` the significant
    wave height at the toe. The Van der Meer formulae are given by:

    .. math::
       \\frac{H_{s}}{\\Delta D_{n 50}}=c_{p l} P^{0.18}\\left(\\frac{S}
       {\\sqrt{N}}\\right)^{0.2} \\xi_{m}^{-0.5} \\; \\;
       \\text{for plunging waves}

       \\frac{H_{s}}{\\Delta D_{n 50}}=c_{s} P^{-0.13}\\left(\\frac{S}
       {\\sqrt{N}}\\right)^{0.2} \\sqrt{\\cot \\alpha} \\, \\xi_{m}^{P}
       \\; \\; \\text{for surging waves}

    As the formulae are based on a large amount of experiments, there
    is a certain reliability for the model constants. For :math:`C_{pl}`
    this is given by :math:`6.2 \\pm 0.4` and :math:`1 \\pm 0.08` for
    :math:`C_{s}`. Furthermore, each parameter has its own range of
    validity. The range of validity of parameters is presented in the
    table below.

    +------------------------------------+---------------------+-------------+
    | Parameter                          |        Symbol       |  Range      |
    +====================================+=====================+=============+
    | Slope angle                        |      tan(alpha)     | 1:6 - 1:1.5 |
    +------------------------------------+---------------------+-------------+
    | Number of waves                    |          N          |   < 7500    |
    +------------------------------------+---------------------+-------------+
    | Fictitious wave steepness          |        s_om         | 0.01 - 0.06 |
    +------------------------------------+---------------------+-------------+
    | Surf similarity parameter          |        xi_m         |  0.7 - 7    |
    +------------------------------------+---------------------+-------------+
    | Relative buoyant density of armour |        Delta        |    1 - 2.1  |
    +------------------------------------+---------------------+-------------+
    | Relative water depth at toe        |      h/H_s-toe      |    > 3      |
    +------------------------------------+---------------------+-------------+
    | Notional permeability parameter    |          P          |  0.1 - 0.6  |
    +------------------------------------+---------------------+-------------+
    | Armour gradation                   |      Dn85/Dn15      |    < 2.5    |
    +------------------------------------+---------------------+-------------+
    | Damage-storm duration ratio        |      Sd/sqrt(N)     |    < 0.9    |
    +------------------------------------+---------------------+-------------+
    | Stability number                   |   Hs/(Delta Dn50)   |    1 - 4    |
    +------------------------------------+---------------------+-------------+
    | Damage level parameter             |         Sd          |    1 - 20   |
    +------------------------------------+---------------------+-------------+

    Parameters
    ----------
    Hs : float
        The significant wave height, :math:`H_{1/3}` of the incident
        waves at the toe [m]
    Delta : float
        Relative buoyant density [-]
    P : float
        Notional permeability of the structure [-]
    Sd : float
        Damage level parameter [-]
    N : int
        Number of incident waves at the toe of the structure [-]
    xi_m : float
        :math:`\\xi_m`, the surf-similarity parameter computed with
        the mean wave period :math:`T_m` [-]
    alpha : float
        Angle of the front slope [rad]
    safety : float, optional, default: 1
        With this parameter the model constants, Cpl and Cs, can be
        decreased to increase the safety. This is implemented as
        :math:`C_{pl} = \\mu - safety \\cdot \\sigma`.

    Returns
    -------
    Dn50 : float
        the nominal diameter of the armourstone [m]
    """
    Cpl = 6.2 - safety*0.4
    Cs = 1 - safety*0.08

    xi_cr = xi_critical(Cpl, Cs, P, alpha)

    if xi_m < xi_cr:  # Plunging waves
        Dn50 = Hs/(Delta*Cpl*P**0.18
                   * (Sd/np.sqrt(N))**0.2*xi_m**-0.5)
    else:  # Surging waves
        Dn50 = Hs/(Delta*Cs*P**-0.13
                   * (Sd/np.sqrt(N))**0.2*np.sqrt(1/np.tan(alpha))
                   * xi_m**P)
    return Dn50


def vandermeer_shallow(Hs, H2, Delta, P, Sd, N, xi_s_min_1, alpha, safety=1):
    """ Van der Meer formulae for shallow water conditions

    Based on the analysis of the stability of rock-armoured slopes in
    many conditions, mainly focussed on conditions with shallow
    foreshores, it was proposed in Van Gent et al (2003) to modify the
    formulae of Van der Meer (1988) to extend its field of application.
    The Van der Meer formulae for shallow water are:

    .. math::
       \\frac{H_{s}}{\\Delta D_{n 50}}=c_{p l} P^{0.18}\\left(\\frac{S}
       {\\sqrt{N}}\\right)^{0.2}\\left(\\frac{H_{s}}{H_{2 \\%}}\\right)
       \\, \\xi_{m-1,0}^{-0.5} \\; \\; \\text{for plunging waves}

       \\frac{H_{s}}{\\Delta D_{n 50}}=c_{s} P^{-0.13}\\left(\\frac{S}
       {\\sqrt{N}}\\right)^{0.2}\\left(\\frac{H_{s}}{H_{2 \\%}}\\right)
       \\sqrt{\\cot \\alpha} \\, \\xi_{m-1,0}^{P} \\; \\;
       \\text{for surging waves}

    As the formulae are based on a large amount of experiments, there
    is a certain reliability for the model constants. For :math:`C_{pl}`
    this is given by :math:`8.4 \\pm 0.7` and :math:`1.3 \\pm 0.15` for
    :math:`C_{s}`. Furthermore, each parameter has its own range of
    validity. The range of validity of parameters is presented in the
    table below.

    +------------------------------------+---------------------+-------------+
    | Parameter                          |        Symbol       |  Range      |
    +====================================+=====================+=============+
    | Slope angle                        |      tan(alpha)     |  1:4 - 1:2  |
    +------------------------------------+---------------------+-------------+
    | Number of waves                    |          N          |   < 3000    |
    +------------------------------------+---------------------+-------------+
    | Fictitious wave steepness          |        s_om         | 0.01 - 0.06 |
    +------------------------------------+---------------------+-------------+
    | Surf similarity parameter          |        xi_m         |    1 - 5    |
    +------------------------------------+---------------------+-------------+
    | Surf similarity parameter          |      xi_m_min_1     |  1.3 - 6.5  |
    +------------------------------------+---------------------+-------------+
    | Wave height ratio                  |       H2%/Hs        |  1.2 - 1.4  |
    +------------------------------------+---------------------+-------------+
    | Deep-water H over h at the toe     |        Hso/h        | 0.25 - 1.5  |
    +------------------------------------+---------------------+-------------+
    | Armour gradation                   |      Dn85/Dn15      |  1.4 - 2.0  |
    +------------------------------------+---------------------+-------------+
    | Core material - armour ratio       |    Dn50-core/Dn50   |    0 - 0.3  |
    +------------------------------------+---------------------+-------------+
    | Stability number                   |   Hs/(Delta Dn50)   |  0.5 - 4.5  |
    +------------------------------------+---------------------+-------------+
    | Damage level parameter             |         Sd          |      < 30   |
    +------------------------------------+---------------------+-------------+


    Parameters
    ----------
    Hs : float
        The significant wave height, :math:`H_{1/3}` of the incident
        waves at the toe [m]
    H2 : float
        :math:`H_{2\\%}`, wave height exceeded by 2% of the incident
        waves at the toe [m]
    Delta : float
        Relative buoyant density [-]
    P : float
        Notional permeability of the structure [-]
    Sd : float
        Damage level parameter [-]
    N : int
        Number of incident waves at the toe of the structure [-]
    xi_m_min_1 : float
        :math:`\\xi_{s-1.0}`, the surf-similarity parameter computed
        with the energy wave period :math:`T_{m-1.0}` [-]
    alpha : float
        Angle of the front slope [rad]
    safety : float, optional, default: 1
        With this parameter the model constants, Cpl and Cs, can be
        decreased to increase the safety. This is implemented as
        :math:`C_{pl} = \\mu - safety \\cdot \\sigma`.

    Returns
    -------
    Dn50 : float
        the nominal diameter of the armourstone [m]
    """
    Cpl = 8.4 - safety*0.7
    Cs = 1.3 - safety*0.15

    xi_cr = xi_critical(Cpl, Cs, P, alpha)

    if xi_s_min_1 < xi_cr: # Plunging waves
        Dn50 = Hs/(Delta*Cpl* P**0.18
                   * (Sd/np.sqrt(N))**0.2 * (Hs/H2) * xi_s_min_1**-0.5)
    else: # Surging waves
        Dn50 = Hs/(1.6*Cs*P**-0.13*(Sd/np.sqrt(N))**0.2
                   * (Hs/H2) * np.sqrt(1/np.tan(alpha))
                   * xi_s_min_1**P)
    return Dn50


def vandermeer(
        LimitState, Delta, P, N, alpha, slope_foreshore, val='max',
        safety=1, logger=None):
    """ Van der Meer formulae for deep and shallow water conditions

    Implementation of Van der Meer formulae for deep and shallow
    water conditions. The function first determines if the water depth
    and wave height are in range of the shallow or deep water formulae,
    and then uses the correct formulae to compute the nominal diameter
    of the armour layer. In case the input is in the range of the deep
    and shallow water formulae both formulae are used to compute the
    diameter, the largest diameter of the two is then returned.

    Parameters
    ----------
    LimitState : :py:class:`LimitState`
        ULS, SLS or another limit state defined with
        :py:class:`LimitState`
    Delta : float
        Relative buoyant density [-]
    P : float
        Notional permeability of the structure [-]
    N : int
        Number of incident waves at the toe of the structure [-]
    alpha : float
        Angle of the front slope [rad]
    slope_foreshore : float
        slope of the foreshore [rad]
    val : {min, max, avg}, optional, default: max
        value to return in case both the deep and shallow water formula
        are valid. min for the lowest value, max for the highest value
        and avg for the average value, default is max.
    safety : float, optional, default: 1
        With this parameter the model constants, Cpl and Cs, can be
        decreased to increase the safety. This is implemented as
        :math:`C_{pl} = \\mu - safety \\cdot \\sigma`.
    logger : dict, optional, default: None
        dict to log messages, must have keys 'INFO' and 'WARNINGS'

    Returns
    -------
    Dn50 : float
        the nominal diameter of the armourstone [m]
    """
    Hs = LimitState.get_Hs('H13')
    H2_per = LimitState.get_H2(slope_foreshore)
    Sd = LimitState['Sd']

    # only the deep water formula is valid
    if LimitState.h/Hs > 3:
        xi_m = LimitState.surf_similarity(alpha=alpha, number='mean')
        Dn50 = vandermeer_deep(
            Hs=Hs, Delta=Delta, P=P, Sd=Sd, N=N, xi_m=xi_m, alpha=alpha,
            safety=safety)
        msg = f'in range of vandermeer_deep with {LimitState.label}'

    # both the deep and shallow water formula are valid
    elif LimitState.h/Hs < 3 and LimitState.h/Hs > 2:
        xi_m = LimitState.surf_similarity(alpha=alpha, number='mean')
        Dn50_d = vandermeer_deep(
            Hs=Hs, Delta=Delta, P=P, Sd=Sd, N=N, xi_m=xi_m, alpha=alpha,
            safety=safety)

        xi_min_1 = LimitState.surf_similarity(alpha=alpha, number='spectral')
        Dn50_s = vandermeer_shallow(
            Hs=Hs, H2=H2_per, Delta=Delta, P=P, Sd=Sd, N=N,
            xi_s_min_1=xi_min_1, alpha=alpha, safety=safety)

        if val == 'max':
            # user wants maximum value
            if Dn50_d >= Dn50_s:
                # deep water formula is normative
                Dn50 = Dn50_d
                formula = 'vandermeer_deep'
            else:
                # shallow water formula is normative
                Dn50 = Dn50_s
                formula = 'vandermeer_shallow'

        elif val == 'min':
            # user wants minimum value
            if Dn50_d >= Dn50_s:
                # shallow water formula is normative
                Dn50 = Dn50_s
                formula = 'vandermeer_shallow'
            else:
                # deep water formula is normative
                Dn50 = Dn50_d
                formula = 'vandermeer_deep'

        elif val == 'avg':
            # compute average
            Dn50 = (Dn50_d + Dn50_s)/2
            formula = 'both'

        else:
            # invalid argument for val
            raise NotSupportedError(
                f'val = {val} is not implemented, must be min, max or avg')


        msg = ('in range of vandermeer_deep and vandermeer_shallow with '
              f'{LimitState.label}, used Dn50 of {formula} as val = {val}')

    # only the shallow water formula is valid
    else:
        xi_min_1 = LimitState.surf_similarity(alpha=alpha, number='spectral')
        Dn50 = vandermeer_shallow(
            Hs=Hs, H2=H2_per, Delta=Delta, P=P, Sd=Sd, N=N,
            xi_s_min_1=xi_min_1, alpha=alpha, safety=safety)
        msg = f'in range of vandermeer_shallow with {LimitState.label}'

    # add msg to log which formula was used
    if logger != None:
        logger['INFO'].append(msg)

    return Dn50


def hudson(H, Kd, Delta, alpha):
    """ Hudson formula for armour layer stability (Hudson, 1959)

    The Hudson formula is based on experimental research on the stability
    of rock and armour units (Tetrapods and Dolose) on a slope. The
    formula is derived by fitting a line through the data points. All
    unknown factors are included in the 'dustbin' factor :math:`k_D`
    (Van den Bos and Verhagen, 2018). The Hudson formula expressed in
    the form with the stability number is:

    .. math::
       \\frac{H}{\\Delta D_{n 50}}=\\sqrt[3]{k_{D} \\cot \\alpha}

    .. warning::
       The formula is based on regular waves (Hudson, 1959), therefore
       the wave height H to use in case of irregular waves is not
       defined. CERC (1984, p. 7-203) advised in the Shore Protection
       Manual to use :math:`H_{1/10}`, the average of the highest 10%
       of the waves.

    .. note::
       Although the Hudson formula is derived for the stability of rock
       and armour units, Van den Bos and Verhagen (2018) recommend to
       use the Van der Meer formula for the stability of rock. As the
       Van der Meer formula includes the effects of storm duration,
       wave period, the structures permeability and a clearly defined
       damage level (CIRIA, CUR, CETMEF, 2007, p. 567).

    Parameters
    ----------
    H : float
        The design wave height [m]
    Kd : int
        Stability coefficient [-]
    Delta : float
        Relative buoyant density [-]
    alpha : float
        Angle of the front slope [rad]

    Returns
    -------
    Dn50 : float
        the nominal diameter [m]
    """
    Dn50 = H/(Delta*(Kd/np.tan(alpha))**(1/3))
    return Dn50



import numpy as np
import pandas as pd
import os
import pickle
from warnings import catch_warnings
from tabulate import tabulate

from .utils._kwarg_validator import _process_kwargs, _RM_vkwargs, _C_vkwargs
from .utils._design_explorer import _DE_params
from .utils.exceptions import RockGradingError, ArmourUnitsError, InputError, user_warning, NotSupportedError
from .utils._progress import ProgressBar
from .utils._excel_utils import _convert_string
from .utils.cost import _process_cost, cost_influence
from .caisson import Caisson
from .conditions import LimitState
from .material import read_grading, read_units
from .rubble import RockRubbleMound, ConcreteRubbleMound


def read_configurations(
        filepath, structure, kd=None, name=None, LS=None, Grading=None,
        ArmourUnits=None, BermMaterial=None):
    """ Conceptual design for multiple breakwaters from an Excel file

    Make a conceptual design for multiple (types) of breakwaters from
    an Excel input file. The Excel input file can be generated with
    :obj:`bw.generate_excel <breakwater.utils.input_generator.generate_excel>`

    .. warning:
       When reading Xbloc or XblocPlus armour units from an Excel file,
       the correction factor applied to the nominal diameter is not
       computed. Specify :py:class:`Xbloc` or :py:class:`XblocPlus`
       for ArmourUnits so that the most optimal design is made.

    Parameters
    ----------
    structure : {'RRM', 'CRM', 'RC', 'CC'}
        structure for which conceptual designs must be generated. RRM
        for a rubble mound with rock as armour layer, CRM for a rubble
        mound with concrete armour units as armour layer, RC for a
        vertical (composite) breakwater with rock as armour layer for
        the foundation and CC for a vertical (composite) breakwater with
        concrete armour units as armour layer for the foundation.
    kd : int, optional, default: None
        stability coefficient [-]
    name : str, optional, default: None
        name of the ArmourUnit
    LS : py:class:`LimitState`, optional, default: None
        ULS, SLS or another limit state defined with
        :py:class:`LimitState`, by default the LimitState is read from
        the Excel input file
    Grading : :py:class:`RockGrading`, optional, default: None
        standard rock grading defined in the NEN-EN 13383-1 or a user
        defined rock grading. By default the Grading is read from
        the Excel input file
    ArmourUnit : obj, optional, default: None
        armour unit class which inherits from :py:class:`ConcreteArmour`,
        for instance :py:class:`Xbloc` or :py:class:`XblocPlus`. By
        default the ArmourUnit is read from the Excel input file. This
        argument is used for RRM.
    BermMaterial : obj, optional, default: None
        armour unit class which inherits from :py:class:`ConcreteArmour`,
        for instance :py:class:`Xbloc` or :py:class:`XblocPlus`. By
        default the BermMaterial is read from the Excel input file.
        This argument is used for CC.

    Returns
    -------
    bw.Configurations
        a Configurations object with breakwater concepts
    """
    # convert the input of structure to a list
    if isinstance(structure, list):
        # must be a list so no change
        structure = structure
    elif isinstance(structure, str):
        # convert single input to list
        structure = [structure]

    # import the excel as a DataFrame
    df = pd.read_excel(filepath, skiprows=1, sheet_name='Parameters')

    # generate input dict
    params = {}

    # load LimitState
    if LS is None:
        # load LimitState data from excel file
        df_ls = pd.read_excel(filepath, index_col=0, sheet_name='LimitState')

        # convert to dict
        input_ls = df_ls.to_dict()['Value']

        # generate LimitState object
        params['LimitState'] = LimitState(**input_ls)
    else:
        params['LimitState'] = LS

    # load RockGrading
    if Grading is None:
        # load Grading
        params['Grading'] = read_grading(filepath, sheet_name='RockGrading')
    else:
        params['Grading'] = Grading

    # load ArmourUnits
    if ArmourUnits is None and 'CRM' in structure:
        # check if kd and name are given
        if kd is None:
            raise InputError(
                ('Missing argument: kd, when reading ArmourUnits from Excel '
                 'the kd value must be given as input'))
        if name is None:
            raise InputError(
                ('Missing argument: name, when reading ArmourUnits from Excel '
                 'the name of the ArmourUnit must be given as input'))

        # load ArmourUnits
        params['ArmourUnit'] = read_units(
            filepath, kd=kd, name=name, sheet_name='ArmourUnit')
    elif ArmourUnits is not None and 'CRM' in structure:
        params['ArmourUnit'] = ArmourUnit

    # load BermMaterial
    if BermMaterial is None and 'CC' in structure:
        # check if kd and name are given
        if kd is None:
            raise InputError(
                ('Missing argument: kd, when reading ArmourUnits from Excel '
                 'the kd value must be given as input'))
        if name is None:
            raise InputError(
                ('Missing argument: name, when reading ArmourUnits from Excel '
                 'the name of the ArmourUnit must be given as input'))

        # load ArmourUnits
        params['BermMaterial'] = read_units(
            filepath, kd=kd, name=name, sheet_name='BermMaterial')
    elif BermMaterial is not None and 'CC' in structure:
        params['BermMaterial'] = BermMaterial

    # set column names and drop validation
    df.columns = ['parameter', 'value', 'min', 'max', 'num', 'validation']
    df.drop(labels=['validation'],axis=1, inplace=True)

    parameters = df['parameter'].values
    values = df['value'].values

    for i, parameter in enumerate(parameters):
        if 'slope' in parameter:
            # check if value is nan
            if not isinstance(values[i], str) and np.isnan(values[i]):
                # parameter is varying
                # convert min and max to tuple
                min = _convert_string(df['min'].values[i], 'slope', parameter)
                max = _convert_string(df['max'].values[i], 'slope', parameter)

                # get num
                num = df['num'].values[i]

                # set value
                value = (min, max, int(num))

            else:
                # parameter is set as constant parameter
                value = _convert_string(values[i], 'slope', parameter)

        elif 'lambda' in parameter:
            # convert lambda to list
            value = _convert_string(values[i], 'lambda', parameter)

        else:
            # get the value
            value = values[i]

            # check if filter rule
            if isinstance(value, str):
                params[parameter] = value
                continue

            # check if value is nan
            if np.isnan(value):
                # get varying input
                min = df['min'].values[i]
                max = df['max'].values[i]
                num = df['num'].values[i]

                # check if one of the values is nan
                if np.isnan([min, max, num]).all():
                    # if all are nan set value to None
                    value = None
                    continue

                elif np.isnan([min, max, num]).any():
                    raise TypeError(
                        (f'Varying parameter {parameter} is incorrectly '
                          'formatted'))

                # convert to tuple
                value = (min, max, int(num))

        # add to params dict
        params[parameter] = value

    return Configurations(structure=structure, **params)

def read_breakwaters(filepath):
    """ Read a breakwaters file into Configurations

    Parameters
    ----------
    filepath : path
        any valid string path is acceptable, filepath must end with
        .breakwaters

    Returns
    -------
    bw.Configurations
        a Configurations object with breakwater concepts
    """
    # get the extension of the file
    extension = os.path.splitext(filepath)[1]

    # check if a extension has been included in the filepath
    if not extension:
        # if not add .breakwaters to the filepath
        filepath = f'{filepath}.breakwaters '
    else:
        # check if extension is .breakwaters
        if extension == '.breakwaters':
            pass
        else:
            raise InputError(
                (f'Can\'t load Configurations from {extension}, must be '
                  '.breakwaters'))

    # load breakwaters file, and return Configurations object
    file = open(filepath, 'rb')
    config = pickle.load(file)

    return config


class Configurations:
    """ Make a conceptual design for multiple breakwaters

    With this class a parametric design of one or more types of
    breakwaters. The types of structures included in the tool are: a
    rubble mound breakwater with rock (RRM) or concrete armour units
    (CRM) as armour layer and a vertical (composite) breakwater with
    rock (RC) or concrete armour units (CC) as armour layer of the
    foundation layer. Each structure has its own set of required
    (table 1) and optional (table 2) parameters.

    For the parametric design some parameters are allowed to vary,
    these parameters have a superscript in table 1 and 2. The input of
    a varying parameter must be a tuple with a length of 3, i.e.
    (min, max, num), where min and max are the minimum and maximum
    value and num is the number of samples to generate.

    .. note::
       Alternatively, a design can also be made from an Excel input file
       by using :py:func:`read_configurations`. The required Excel input
       file can be generated by using
       :obj:`bw.generate_excel <breakwater.utils.input_generator.generate_excel>`

    Table 1: required parameters

    +--------------------+-----+-----+-----+-----+
    | Parameter          | RRM | CRM | RC  | CC  |
    +====================+=====+=====+=====+=====+
    | LimitState         |  x  |  x  |  x  |  x  |
    +--------------------+-----+-----+-----+-----+
    | rho_w              |  x  |  x  |  x  |  x  |
    +--------------------+-----+-----+-----+-----+
    | slope_foreshore    |  x  |  x  |  x  |  x  |
    +--------------------+-----+-----+-----+-----+
    | Grading            |  x  |  x  |  x  |  x  |
    +--------------------+-----+-----+-----+-----+
    | slope :sup:`1`     |  x  |  x  |     |     |
    +--------------------+-----+-----+-----+-----+
    | B :sup:`1`         |  x  |  x  |     |     |
    +--------------------+-----+-----+-----+-----+
    | Dn50_core :sup:`1` |  x  |  x  |     |     |
    +--------------------+-----+-----+-----+-----+
    | N                  |  x  |     |     |     |
    +--------------------+-----+-----+-----+-----+
    | ArmourUnit         |     |  x  |     |     |
    +--------------------+-----+-----+-----+-----+
    | Pc :sup:`1`        |     |     |  x  |  x  |
    +--------------------+-----+-----+-----+-----+
    | rho_c              |     |     |  x  |  x  |
    +--------------------+-----+-----+-----+-----+
    | rho_fill           |     |     |  x  |  x  |
    +--------------------+-----+-----+-----+-----+
    | Bm :sup:`1`        |     |     |  x  |  x  |
    +--------------------+-----+-----+-----+-----+
    | hb :sup:`1`        |     |     |  x  |  x  |
    +--------------------+-----+-----+-----+-----+
    | mu                 |     |     |  x  |  x  |
    +--------------------+-----+-----+-----+-----+
    | BermMaterial       |     |     |     |  x  |
    +--------------------+-----+-----+-----+-----+

    :sup:`1` Parameter is allowed to vary, enter as a tuple with
    (min, max, num), where min and max are the minimum and maximum
    value and num is the number of samples to generate.

    Table 2: Optional parameters with default values

    +---------------------------+---------+-----+-----+-----+-----+
    | Parameter                 | Default | RRM | CRM | RC  | CC  |
    +===========================+=========+=====+=====+=====+=====+
    | safety                    |    1    |  o  |  o  |  o  |  o  |
    +---------------------------+---------+-----+-----+-----+-----+
    | beta                      |    0    |  o  |  o  |  o  |  o  |
    +---------------------------+---------+-----+-----+-----+-----+
    | Soil                      |  None   |  o  |  o  |  o  |  o  |
    +---------------------------+---------+-----+-----+-----+-----+
    | slope_toe :sup:`1`        |  (2,3)  |  o  |  o  |     |     |
    +---------------------------+---------+-----+-----+-----+-----+
    | phi                       |    40   |  o  |  o  |     |     |
    +---------------------------+---------+-----+-----+-----+-----+
    | B_toe :sup:`1`            |  None   |  o  |  o  |     |     |
    +---------------------------+---------+-----+-----+-----+-----+
    | vdm                       |   max   |  o  |     |     |     |
    +---------------------------+---------+-----+-----+-----+-----+
    | layers_rock               |    2    |  o  |     |  o  |     |
    +---------------------------+---------+-----+-----+-----+-----+
    | layers_units              |    1    |     |  o  |     |  o  |
    +---------------------------+---------+-----+-----+-----+-----+
    | layers_underlayer         |    2    |  o  |  o  |     |     |
    +---------------------------+---------+-----+-----+-----+-----+
    | filter_rule               |  None   |     |  o  |     |  o  |
    +---------------------------+---------+-----+-----+-----+-----+
    | pe_max                    |   500   |     |     |  o  |  o  |
    +---------------------------+---------+-----+-----+-----+-----+
    | SF_sliding                |   1.2   |     |     |  o  |  o  |
    +---------------------------+---------+-----+-----+-----+-----+
    | SF_turning                |   1.2   |     |     |  o  |  o  |
    +---------------------------+---------+-----+-----+-----+-----+
    | slope_foundation :sup:`1` |  (2,3)  |     |     |  o  |  o  |
    +---------------------------+---------+-----+-----+-----+-----+
    | \lambda_                  | [1,1,1] |     |     |  o  |  o  |
    +---------------------------+---------+-----+-----+-----+-----+


    :sup:`1` Parameter is allowed to vary, enter as a tuple with
    (min, max, num), where min and max are the minimum and maximum
    value and num is the number of samples to generate.

    Parameters
    ----------
    structure : {'RRM', 'CRM', 'RC', 'CC'}
        structure for which conceptual designs must be generated. RRM
        for a rubble mound with rock as armour layer, CRM for a rubble
        mound with concrete armour units as armour layer, RC for a
        vertical (composite) breakwater with rock as armour layer for
        the foundation and CC for a vertical (composite) breakwater with
        concrete armour units as armour layer for the foundation.
    LimitState : :py:class:`LimitState` or list of :py:class:`LimitState`
        ULS, SLS or another limit state defined with
        :py:class:`LimitState`
    rho_w : float
        density of water [kg/m³]
    slope_foreshore : tuple
        slope of the foreshore (V, H). For example a slope of 1:100 is
        defined as (1, 100)
    Grading : :py:class:`RockGrading`
        standard rock grading defined in the NEN-EN 13383-1 or a user
        defined rock grading.
    safety : float, optional, default: 1
        safety factor of design (number of standard deviations from the
        mean)

    Keyword arguments
    -----------------
    slope : tuple
        Slope of the armour layer (V,H). For example a slope of 3V:4H
        is defined as (3,4). Required for RRM and CRM.
    B : float
        Crest width [m], required for RRM and CRM
    Dn50_core : float
        nominal diameter for the stones in the core of the breakwater [m],
        required for RRM and CRM
    N : int
        Number of incident waves at the toe of the structure [-],
        required for RRM
    ArmourUnit : obj
        armour unit class which inherits from :py:class:`ConcreteArmour`,
        for instance :py:class:`Xbloc` or :py:class:`XblocPlus`. Required
        argument for CRM.
    Pc : float
        contribution of concrete to the total mass of the caisson.
        value between 0 and 1. Required argument for RC and CC.
    rho_c : float
        density of concrete [kg/m³], required argument for RC and CC.
    rho_fill : float
        density of the fill material [kg/m³], required argument for RC
        and CC.
    Bm : float
        width of the berm [m], required argument for RC and CC.
    hb : float
        height of the foundation layer [m], required argument for RC
        and CC.
    mu : float
        friction factor between the caisson and the foundation [-],
        required argument for RC and CC.
    BermMaterial : obj
        should be a :py:class:`RockGrading` or armour unit class which
        inherits from :py:class:`ConcreteArmour`, for instance
        :py:class:`Xbloc` or :py:class:`XblocPlus`. Required argument
        for CC
    beta : float, optional, default: 0
        angle between direction of wave approach and a line normal to
        the breakwater (degrees). Optional argument for RC and CC,
        default value is 0.
    slope_toe : tuple, optional, default: (2,3)
        Slope of the armour layer (V,H). For example a slope of 2V:3H
        is defined as (2,3). Optional argument for RRM and CRM, default
        value is (2,3)
    B_toe : float, optional, default: None
        width of the top of the toe in meters. By default the width of
        toe is taken as 3 * Dn50_toe. Optional argument for RRM and
        CRM.
    vdm : {min, max, avg}, optional, default: max
        value to return in case both the deep and shallow water formula
        are valid. min for the lowest value, max for the highest value
        and avg for the average value, default is max. Optional argument
        for RRM.
    layers_rock : int, optional, default: 2
        number of layer in the armour layer for a rubble mound breakwater
        with rock as armour layer. Optional argument for RRM and RC,
        default value is 2
    layers_units : int, optional, default: 1
        number of layer in the armour layer for a rubble mound
        breakwater with concrete armour units as armour layer. Optional
        argument for RRM and CC, default value is 1
    layers_underlayer : int, optional, default: 2
        number of layers in the underlayer. Optional argument for RRM
        and CRM
    filter_rule : {'Rock', 'Xbloc', 'XblocPlus'}, optional, default: None
        filter rule to use for the substructure of the breakwater, for
        Rock, Xbloc and XblocPlus the correct filter rule is
        automatically selected. In case another type of armour layer is
        used one of these filter rules must be chosen. Optional argument
        for CRM, required if armour unit is not Xbloc or XblocPlus,
        default value is None
    pe_max : float, optional, default: 500
        maximum value of the bearing pressure at the heel of the caisson.
        Default value is set to 500 kPa, Goda (2000) advises a value
        between 400 and 500 kPa. Optional parameter for RC and CC
    SF_sliding : float, optional, default: 1.2
        safety factor against sliding. Default value according to
        Goda (2000). Optional argument for RC and CC, default value is 1.2
    SF_turning : float, optional, default: 1.2
        safety factor against sliding. Default value according to
        Goda (2000). Optional argument for RC and CC, default value is 1.2
    slope_foundation : tuple, optional, default: (2,3)
        Slope of the armour layer (V,H). For example a slope of 2V:3H
        is defined as (2,3). Optional argument for RC and CC, default
        is (2,3)
    lambda_ : list, optional, default: [1,1,1]
        modification factors of Takahasi (2002) for alternative
        monolithic breakwater. Input must be
        \\lambda_= [:math:`\\lambda_1, \\lambda_2, \\lambda_3`].
        Optional argument for RC and CC, default value is [1,1,1]
    Soil : :py:class:`Soil`, optional, default: None
        by default Soil is None, which means that the geotechnical checks
        are not performed. By specifying a Soil object, the geotechnical
        checks are automatically performed. Optional argument for all
        structures
    phi : float, optional, default: 40
        internal friction angle of rock [degrees]. Optional argument for
        RRM and CRM.

    Attributes
    ----------
    df : pd.DataFrame
        DataFrame with all generated concepts
    """
    def __init__(
            self, structure, LimitState, rho_w, slope_foreshore, Grading,
            Soil=None, safety=1, **kwargs):
        """ See help(Configurations) for more info """
        # add LimitState, rho_w and slope_foreshore to kwargs input
        # so that input can be validated
        kwargs['LimitState'] = LimitState
        kwargs['rho_w'] = rho_w
        kwargs['slope_foreshore'] = slope_foreshore
        kwargs['safety'] = safety
        kwargs['Grading'] = Grading
        kwargs['Soil'] = Soil

        # set grading as attribute for adding cost
        self._Grading = Grading

        # convert the input of structure to a list
        if isinstance(structure, list):
            # must be a list so no change
            structure = structure
        elif isinstance(structure, str):
            # convert single input to list
            structure = [structure]

        # set empty configs and vkwargs
        RM_config = {}
        RM_vkwargs = {}
        C_config = {}
        C_vkwargs = {}

        # check if the input is correct for the type of structure
        # and set default values if a default value must be set
        if 'RRM' in structure and 'CRM' in structure:
            RRM_compute = True
            CRM_compute = True

            RM_vkwargs = _RM_vkwargs(type='both')
            RM_config = _process_kwargs(kwargs=kwargs, vkwargs=RM_vkwargs)

            # unpack constant arguments for both
            beta = RM_config['beta']
            phi = RM_config['phi']

            # unpack constant arguments for RRM
            N = RM_config['N']
            layers_rock = RM_config['layers_rock']
            vdm = RM_config['vdm']

            # unpack constant arguments for CRM
            ArmourUnit = RM_config['ArmourUnit']
            layers_units = RM_config['layers_units']
            filter_rule = RM_config['filter_rule']

        elif 'RRM' in structure:
            RRM_compute = True
            CRM_compute = False

            RM_vkwargs = _RM_vkwargs(type='Rock')
            RM_config = _process_kwargs(kwargs=kwargs, vkwargs=RM_vkwargs)

            # unpack constant arguments for RRM
            beta = RM_config['beta']
            N = RM_config['N']
            layers_rock = RM_config['layers_rock']
            vdm = RM_config['vdm']
            phi = RM_config['phi']

        elif 'CRM' in structure:
            RRM_compute = False
            CRM_compute = True

            RM_vkwargs = _RM_vkwargs(type='ArmourUnit')
            RM_config = _process_kwargs(kwargs=kwargs, vkwargs=RM_vkwargs)

            # unpack constant arguments for CRM
            beta = RM_config['beta']
            ArmourUnit = RM_config['ArmourUnit']
            layers_units = RM_config['layers_units']
            filter_rule = RM_config['filter_rule']
            phi = RM_config['phi']

        elif 'RRM' not in structure and 'CRM' not in structure:
            RRM_compute = False
            CRM_compute = False

        # input for caisson in structure
        if 'RC' in structure and 'CC' in structure:
            RC_compute = True
            CC_compute = True

            C_vkwargs = _C_vkwargs(type='both')
            C_config = _process_kwargs(kwargs=kwargs, vkwargs=C_vkwargs)

            # unpack constant arguments for both
            rho_c = C_config['rho_c']
            rho_fill = C_config['rho_fill']
            mu = C_config['mu']
            SF_sliding = C_config['SF_sliding']
            SF_turning = C_config['SF_turning']
            beta = C_config['beta']
            lambda_ = C_config['lambda_']
            filter_rule = C_config['filter_rule']
            pe_max = C_config['pe_max']

            # unpack constant arguments for RC
            layers_rock = C_config['layers_rock']

            # unpack constant arguments for CC
            layers_units = C_config['layers_units']
            BermMaterial = C_config['BermMaterial']

        elif 'RC' in structure:
            RC_compute = True
            CC_compute = False

            C_vkwargs = _C_vkwargs(type='Rock')
            C_config = _process_kwargs(kwargs=kwargs, vkwargs=C_vkwargs)

            # unpack constant arguments for both
            rho_c = C_config['rho_c']
            rho_fill = C_config['rho_fill']
            mu = C_config['mu']
            SF_sliding = C_config['SF_sliding']
            SF_turning = C_config['SF_turning']
            beta = C_config['beta']
            lambda_ = C_config['lambda_']
            filter_rule = C_config['filter_rule']
            pe_max = C_config['pe_max']

            # unpack constant arguments for RC
            layers_rock = C_config['layers_rock']

        elif 'CC' in structure:
            RC_compute = False
            CC_compute = True

            C_vkwargs = _C_vkwargs(type='ArmourUnit')
            C_config = _process_kwargs(kwargs=kwargs, vkwargs=C_vkwargs)

            # unpack constant arguments for both
            rho_c = C_config['rho_c']
            rho_fill = C_config['rho_fill']
            mu = C_config['mu']
            SF_sliding = C_config['SF_sliding']
            SF_turning = C_config['SF_turning']
            beta = C_config['beta']
            lambda_ = C_config['lambda_']
            filter_rule = C_config['filter_rule']
            pe_max = C_config['pe_max']

            # unpack constant arguments for CC
            layers_units = C_config['layers_units']
            BermMaterial = C_config['BermMaterial']

        elif 'RC' not in structure and 'CC' not in structure:
            RC_compute = False
            CC_compute = False

        # check if a computation must be made
        if (not RRM_compute and not CRM_compute
                and not RC_compute and not CC_compute):
            # noting to compute, raise error
            raise NotSupportedError(
                (f'{structure} has not been implemented, supported structures'
                  ' are \'RRM\', \'CRM\', \'RC\' and \'CC\''))

        # initialize a df for to store the _get_concept_set
        self.df = pd.DataFrame()

        # design all concepts for a rubble mound breakwater
        # get all possible combinations of the varying arguments
        RM_varying, RM_num_combinations = self._get_combinations(
            vkwargs=RM_vkwargs, config=RM_config)

        for i in range(RM_num_combinations):
            # in first iteration check if computation is needed
            # and unpack arguments if needed
            if i == 0:
                if not RRM_compute and not CRM_compute:
                    # break loop if computation is not needed
                    break

                if RRM_compute and CRM_compute:
                    # both thus double the number of computations
                    num = 2*RM_num_combinations
                else:
                    # only design 1 structure
                    num = RM_num_combinations

                # initialize progress bar for rubble mound computations
                RM_bar = ProgressBar(
                    number=num, task='Computing Rubble Mound')

            # set id of bw
            id = i + 1

            # get the current concept
            concept = self._get_concept_set(configs=RM_varying, id=id)

            # unpack varying arguments same for RRM and CRM
            B = concept['B']
            Dn50_core = concept['Dn50_core']
            B_toe = concept['B_toe']
            slope = concept['slope']
            slope_toe = concept['slope_toe']

            if RRM_compute:

                try:
                    with catch_warnings(record=True) as w:
                        RM_rock = RockRubbleMound(
                            slope=slope, slope_foreshore=slope_foreshore,
                            rho_w=rho_w, B=B, N=N, LimitState=LimitState,
                            Grading=Grading, Dn50_core=Dn50_core,
                            safety=safety, slope_toe=slope_toe, B_toe=B_toe,
                            layers=layers_rock, vdm=vdm, Soil=Soil, phi=phi,
                            id=id)

                except RockGradingError:
                    # if there is no rock class in the grading that
                    # can satisfy the computed Dn50 of the armour layer
                    RM_rock = None

                # save the concept to a temporary df and append to df
                temp_df = pd.DataFrame(data={'type': ['RRM'],
                                             'id': [id],
                                             'concept': [RM_rock],
                                             'B': [B],
                                             'Dn50_core': [Dn50_core],
                                             'B_toe': [B_toe],
                                             'slope': [slope],
                                             'slope_toe': [slope_toe],
                                             'warnings': [w]})
                self.df = self.df.append(temp_df, ignore_index=True, sort=True)

                RM_bar.next()

            if CRM_compute:
                try:
                    with catch_warnings(record=True) as w:
                        RM_units = ConcreteRubbleMound(
                            slope=slope, slope_foreshore=slope_foreshore, B=B,
                            rho_w=rho_w, LimitState=LimitState, safety=safety,
                            Grading=Grading, ArmourUnit=ArmourUnit, phi=phi,
                            Dn50_core=Dn50_core, slope_toe=slope_toe,
                            B_toe=B_toe, layers=layers_units, Soil=Soil, id=id,
                            filter_rule=filter_rule)

                except ArmourUnitsError:
                    # if there is no class of armour unit that
                    # can satisfy the computed Dn50 of the armour layer
                    units_bw = None

                # save the concept to a temporary df and append to df
                temp_df = pd.DataFrame(data={'type': ['CRM'],
                                             'id': [id],
                                             'concept': [RM_units],
                                             'B': [B],
                                             'Dn50_core': [Dn50_core],
                                             'B_toe': [B_toe],
                                             'slope': [slope],
                                             'slope_toe': [slope_toe],
                                             'warnings': [w]})
                self.df = self.df.append(temp_df, ignore_index=True, sort=True)

                RM_bar.next()

            if id == RM_num_combinations:
                RM_bar.finish()

        # design all concepts for a caisson breakwater
        # get all possible combinations of the varying arguments
        C_varying, C_num_combinations = self._get_combinations(
            vkwargs=C_vkwargs, config=C_config)

        for i in range(C_num_combinations):
            # in first iteration check if computation is needed
            # and unpack arguments if needed
            if i == 0:
                if not RC_compute and not CC_compute:
                    # break loop if computation is not needed
                    break

                if RC_compute and CC_compute:
                    # both thus double the number of computations
                    num = 2*C_num_combinations
                else:
                    # only design 1 structure
                    num = C_num_combinations

                if RRM_compute or CRM_compute:
                    # set task name for nice allignment with RM
                    task_name = 'Computing Caisson     '
                else:
                    # normal task name
                    task_name = 'Computing Caisson'

                # initialize progress bar for caisson computations
                C_bar = ProgressBar(number=num, task=task_name)

            # set id of bw
            id = i + 1

            # get the current concept
            concept = self._get_concept_set(configs=C_varying, id=id)

            # unpack varying arguments
            Pc = concept['Pc']
            Bm = concept['Bm']
            hb = concept['hb']
            slope_foundation = concept['slope_foundation']

            if RC_compute:
                # design with rock as armour layer for the foundation
                try:
                    with catch_warnings(record=True) as w:
                        C_rock = Caisson(
                            Pc=Pc, rho_c=rho_c, rho_fill=rho_fill, beta=beta,
                            rho_w=rho_w, Bm=Bm, hb=hb, layers=layers_rock,
                            BermMaterial=Grading, mu=mu, LimitState=LimitState,
                            safety=safety, slope_foreshore=slope_foreshore,
                            SF_sliding=SF_sliding, SF_turning=SF_turning,
                            slope_foundation=slope_foundation, lambda_=lambda_,
                            filter_rule=filter_rule, Grading=Grading, id=id,
                            pe_max=pe_max, Soil=Soil)

                except RockGradingError:
                    # if there is no rock class in the grading that
                    # can satisfy the computed Dn50 of the armour layer
                    C_rock = None

                # save the concept to a temporary df and append to df
                temp_df = pd.DataFrame(data={'type': ['RC'],
                                             'id': [id],
                                             'concept': [C_rock],
                                             'Pc': [Pc],
                                             'Bm': [Bm],
                                             'hb': [hb],
                                             'slope_foundation': [slope_foundation],
                                             'warnings': [w]})
                self.df = self.df.append(temp_df, ignore_index=True, sort=True)

                C_bar.next()

            if CC_compute:

                try:
                    with catch_warnings(record=True) as w:
                        C_units = Caisson(
                            Pc=Pc, rho_c=rho_c, rho_fill=rho_fill, rho_w=rho_w,
                            Bm=Bm, hb=hb, layers=layers_units, mu=mu, beta=beta,
                            BermMaterial=BermMaterial, LimitState=LimitState,
                            slope_foreshore=slope_foreshore, safety=safety,
                            SF_sliding=SF_sliding, SF_turning=SF_turning,
                            slope_foundation=slope_foundation, lambda_=lambda_,
                            filter_rule=filter_rule, Grading=Grading, id=id,
                            pe_max=pe_max, Soil=Soil)

                except ArmourUnitsError:
                    # if there is no class of armour unit that
                    # can satisfy the computed Dn50 of the armour layer
                    C_units = None

                # save the concept to a temporary df and append to df
                temp_df = pd.DataFrame(data={'type': ['CC'],
                                             'id': [id],
                                             'concept': [C_units],
                                             'Pc': [Pc],
                                             'Bm': [Bm],
                                             'hb': [hb],
                                             'warnings': [w]})
                self.df = self.df.append(temp_df, ignore_index=True, sort=True)

                C_bar.next()

            if id == C_num_combinations:
                C_bar.finish()

    @staticmethod
    def _get_concept_set(configs, id):
        """ get unique set of parameters and values """
        index = id - 1

        unique_set = {}

        for param, val in configs.items():
            unique_set[param] = val[index]

        return unique_set

    @staticmethod
    def _get_combinations(vkwargs, config):
        # get the varying and constant parameters
        varying = {}
        fixed = {}

        for param, val in config.items():
            if vkwargs[param]['Constant']:
                # constant parameter, already in RM_config
                continue
            else:
                # varying parameter
                if isinstance(val, np.ndarray):
                    # also given as varying parameter
                    varying[param] = val
                else:
                    # not set as varying thus constant
                    fixed[param] = val

        # create all possible combinations of the varying parameters
        # get the parameters and number of parameters
        parameters = list(varying.keys())
        num_parameters = len(parameters)

        # empty dict to store the parameter set of each concept
        configs = {}

        # create value_index with the index of the value in the list of the dict
        # value_index with 0 is the first concept
        value_index = [0] * num_parameters
        first_concept = True
        current_combination = 0

        while True:
            # check for exit condition
            # first compute maximum number of combinations
            if first_concept:
                max_combinations = 1
                for param, values in varying.items():
                    combinations = len(values)
                    max_combinations = max_combinations * combinations

            if current_combination == max_combinations:
                break

            # saving the concept in configs
            for i in range(num_parameters):
                current_parameter = parameters[i]
                if first_concept:
                    # first time a list for each key must be made
                    configs[current_parameter] = []
                # save the value of the current key
                value = varying[current_parameter][value_index[i]]
                configs[current_parameter].append(value)

            # first concept has been generated so set to False
            first_concept = False


            # set loop variables, current_column_index starts from the right
            change = True
            current_column_index = num_parameters - 1

            while change and current_column_index >= 0:
                # get the length of the current parameter (how many values)
                current_parameter = parameters[current_column_index]
                max_values = len(varying[current_parameter])

                # check if all variants for this level have been made
                if (value_index[current_column_index] + 1) > max_values-1:
                    # all variants for the level have been added
                    # so set the last index to zero
                    value_index[current_column_index] = 0

                    # Change the upper variable by one
                    # We need to increment the immediate upper level loop by one
                    change = True
                else:
                    # add one to the index for the next combination
                    value_index[current_column_index] += 1

                    # set the change to False so that the loop stops for the
                    # current level
                    change = False

                # move one column to the left
                current_column_index -= 1

            current_combination += 1

        # check if there are fixed values
        if fixed:
            # add the fixed parameters to the configs
            for param, val in fixed.items():
                values = [val for i in range(max_combinations)]
                configs[param] = values

        # return the combinations and number of combinations
        return configs, max_combinations

    def add_cost(
            self, core_price=None, unit_price=None, concrete_price=None,
            fill_price=None, transport_cost=None, investment=None,
            length=None):
        """ Compute the cost of each concept

        Compute the cost of each concept and add the cost to the
        :py:attr:`df`. The cost of the rocks must be specified in the
        RockGrading. If transport cost are not included in the price of
        rocks or core_price it can be given with the argument
        transport_cost. For a Caisson breakwater it is possible to
        specify the investment for renting a dry dock, the investment
        is divided through the length of the breakwater to get the
        investment cost per meter.

        .. note::
           The transport_cost are not added to the price of the armour
           layer. The assumption has been made that the cost of
           producing and transporting the armour units is included in
           the unit_price.

        Parameters
        ----------
        core_price : float, optional, default: None
            cost of the core material per m³, required for RRM and CRM
        unit_price : float, optional, default: None
            the cost of an armour unit per m³, required for CRM and CC
        concrete_price : float, optional, default: None
            price of concrete per m³, required for RC and CC
        fill_price : float, optional, default: None
            price of the fill material per m³, required for RC and CC
        transport_cost : float, optional, default: None
            the cost to transport a m³ of rock from the quarry to the
            project location
        investment : float
            the investment required to rent a dry dock
        length : float
            length of the breakwater [m]
        """
        # make dict of the cost for validation
        cost = {
            'core_price': core_price,
            'unit_price': unit_price,
            'concrete_price': concrete_price,
            'fill_price': fill_price}

        # check if all required cost have been given
        for type in self.df.type.unique():
            # validate cost
            _process_cost(type, cost, self._Grading)

        # set list to store cost in
        computed_cost = []

        # iterate over the generated concepts
        for i, row in self.df.iterrows():
            # check if concept is not None
            if row.concept is None:
                # not a valid concept
                computed_cost.append(None)

            else:
                # valid concept
                # check types and compute price
                if row.type == 'RRM':
                    price = row.concept.cost(
                        *row.concept.variantIDs, core_price=core_price,
                        transport_cost=transport_cost, output='variant')

                elif row.type == 'CRM':
                    price = row.concept.cost(
                        *row.concept.variantIDs, core_price=core_price,
                        unit_price=unit_price, transport_cost=transport_cost,
                        output='variant')

                elif row.type == 'RC' or row.type == 'CC':
                    # check if investment cost must be added
                    if investment is not None and length is not None:
                        # add investment cost
                        row.concept.dry_dock(investment, length)

                    price = row.concept.cost(
                        *row.concept.variantIDs, concrete_price=concrete_price,
                        fill_price=fill_price, unit_price=unit_price)

                else:
                    raise NotSupportedError(f'{row.type} is not supported')

                # add cost to list
                computed_cost.append(price)

        # add column to the df
        self.df['cost'] = computed_cost

    def to_design_explorer(
            self, params, mkdir='DesignExplorer', slopes='angles',
            merge_Bm=True, merge_slope_toe=True):
        """ Export concepts to Design Explorer 2

        Creates a folder that can be used in Design Explorer 2, the
        folder consists of the cross sections of all concepts and a csv
        file with the data of the concepts. Parameters supported for
        export can be seen in table 3. To use the folder in Design
        Explorer 2, follow these steps:

        - Upload the folder to your Google Drive
        - Share the folder and get the shareable link
        - Go to http://tt-acm.github.io/DesignExplorer/
        - Click on *Get Data* and paste the shareable link below: *From
          the cloud*
        - Click on *Load Data*
        - Enjoy exploring!

        Table 3: possible parameter to export

        +--------------------------+------------+------------+
        | Parameter                | RRM + CRM  |  RC + CC   |
        +==========================+============+============+
        | cost                     |     o      |     o      |
        +--------------------------+------------+------------+
        | B                        |     o      |     o      |
        +--------------------------+------------+------------+
        | Rc                       |     o      |     o      |
        +--------------------------+------------+------------+
        | computed Dn50 armour     |     o      |     o      |
        +--------------------------+------------+------------+
        | class armour             |     o      |     o      |
        +--------------------------+------------+------------+
        | class Dn50 armour        |     o      |     o      |
        +--------------------------+------------+------------+
        | computed Dn50 underlayer |     o      |     o      |
        +--------------------------+------------+------------+
        | class underlayer         |     o      |     o      |
        +--------------------------+------------+------------+
        | class Dn50 underlayer    |     o      |     o      |
        +--------------------------+------------+------------+
        | slope                    |     o      |            |
        +--------------------------+------------+------------+
        | slope_toe                |     o      | o :sup:`1` |
        +--------------------------+------------+------------+
        | Dn50_core                |     o      |            |
        +--------------------------+------------+------------+
        | B_toe                    |     o      |            |
        +--------------------------+------------+------------+
        | computed Dn50 filter     |     o      |            |
        +--------------------------+------------+------------+
        | class filter             |     o      |            |
        +--------------------------+------------+------------+
        | class Dn50 filter        |     o      |            |
        +--------------------------+------------+------------+
        | Pc                       |            |     o      |
        +--------------------------+------------+------------+
        | hb                       |            |     o      |
        +--------------------------+------------+------------+
        | h_acc                    |            |     o      |
        +--------------------------+------------+------------+
        | Bm                       | o :sup:`2` |     o      |
        +--------------------------+------------+------------+
        | slope_foundation         |            |     o      |
        +--------------------------+------------+------------+
        | UC :sup:`3`              |            |     o      |
        +--------------------------+------------+------------+

        | :sup:`1` slope_foundation is interpreted as slope_toe if
          merge_slope_toe is set to True
        | :sup:`2` B_toe is interpreted as Bm if merge_Bm is set to True
        | :sup:`3` UC is the unity check for the bearing capacity of the
          subsoil

        Parameters
        ----------
        params : list
            list of parameters for the design explorer, parameter must
            be str. See table 3 for the parameters that can be exported/
        mkdir : str, optional, default: concepts
            creates a folder in which all cross sections are saved,
            upload the created folder to your Google Drive to use
            Design Explorer 2
        slopes : {angles, tuples}, optional, default: angles
            how the slopes must be exported. tuples will export the
            slope as (V,H) and angles as an angle in degrees
        merge_Bm : bool, optional, default: True
            True if Bm must be merged with B_toe when a Rubble Mound and
            Vertical breakwater have been designed, False if you do not
            want to merge the columns.
        merge_slope_toe : bool, optional, default: True
            True if slope_foundation must be merged with slope_toe when
            a Rubble Mound and Vertical breakwater have been designed,
            False if you do not want to merge the columns.

        Raises
        ------
        KeyError
            if the cost are asked to export but not yet specified
        """
        # set empty df for export
        to_export = pd.DataFrame()

        # check if RRM and CRM are in structure
        designed_structures = self.df.type.unique()
        if 'CRM' and 'RRM' in designed_structures:
            format_class_as_string = True
        else:
            format_class_as_string = False

        # set dir int variable for generating a new dir
        dir_int = 0

        # while loop to create dir with unique name
        while True:
            # set name of the new_dir
            if dir_int == 0:
                # first iteration, so use specified name
                new_dir = mkdir
            else:
                # specified name already exists so add int to name of dir
                new_dir = f'{mkdir} {dir_int}'

            # check if dir already exists
            if os.path.exists(new_dir):
                # dir already exists, check if it has files
                if len(os.listdir(new_dir)) == 0:
                    # no files thus data can be added, break loop
                    break
                else:
                    # increase dir_int with 1 to make new dir
                    dir_int += 1
            else:
                # dir does not exists, so make one and break
                os.mkdir(new_dir)
                break

        # check if the name of dir is different from the specified
        if mkdir is not new_dir:
            # name different, print user warning to notify user
            user_warning(
                (f'directory {mkdir} already exists, therefore a new directory'
                 f' has been made. The files can be found in {new_dir}'))

        # set up progress bar, and CaseNo
        num_concepts = self.df.shape[0]
        bar = ProgressBar(number=num_concepts, task='Saving')
        CaseNo = 1

        # add list to store all CaseNo in
        all_CaseNo = []

        # start the export
        for index, row in self.df.iterrows():
            # check if concept exists
            if row.concept is None:
                # go to next concept
                bar.next()
                all_CaseNo.append([CaseNo])
                continue

            # compute number of variants for this concept
            num_concepts = len(row.concept.variantIDs)

            # store CaseNo of current concept in a list
            temp_CaseNo = []

            for id in row.concept.variantIDs:
                # save name of the cross section
                file_name = f'{row.type}.{row.id}'

                if num_concepts > 1:
                    # add id if more than 1 concept
                    file_name = f'{file_name}{id}'

                save_name = f'{new_dir}/{file_name}'

                # save cross section of the current concept
                row.concept.plot(id, save_name=save_name)

                data = {'CaseNo': [CaseNo], 'type': [row.type]}

                # get the values from the variant and add to data
                variant = row.concept.get_variant(variantID=id)
                to_explorer = _DE_params(
                    args=params, variant=variant, row=row, concept=row.concept,
                    structure=row.type, slopes=slopes,
                    change_CRM_class=format_class_as_string)
                data.update(to_explorer)

                # check if cost must be included
                if 'cost' in params:
                    # check if cost have been added
                    if 'cost' in self.df.columns:
                        # add cost to data
                        data['cost'] = row.cost[id]
                    else:
                        raise KeyError(
                            'Cost have not been added, use add_cost to add '
                            'cost to the df')

                # add image to data
                data['img'] = [f'{file_name}.png']

                # create a df of data and add to the df to_export
                export_row = pd.DataFrame(data=data)
                to_export = to_export.append(
                    export_row, ignore_index=True, sort=True)

                # add CaseNo to temp CaseNo and increase with 1
                temp_CaseNo.append(CaseNo)
                CaseNo += 1

            # add all stored CaseNo of the current concept in a list
            all_CaseNo.append(temp_CaseNo)

            bar.next()

        bar.finish()

        # add all CaseNo to the df so that concept can be selected by CaseNo
        self.df['CaseNo'] = all_CaseNo

        # post process for to_export
        # if CRM and RRM for class armour values must be sorted
        if format_class_as_string:
            to_export.sort_values(by=['type'], inplace=True)

        # check if all given params are in the columns
        columns = list(to_export)
        not_used_params = np.setdiff1d(
            params, columns, assume_unique=True).tolist()
        if any(not_used_params):
            # remove parameters not used from params
            column_names = [x for x in params if x not in not_used_params]

            # raise UserWarning that not all given params have been exported
            skipped = ', '.join(not_used_params)
            user_warning(
                f'Not all params have been exported, skipped: {skipped}')
        else:
            # all given params have been used
            column_names = params

        # add CaseNo and type as the left most column_names
        # and make img the right most column_names
        column_names[0:0] = ['CaseNo', 'type']
        column_names.insert(len(column_names), 'img')

        # restructure df with order of column_names
        to_export = to_export.loc[:, column_names]

        # merge B_toe with Bm if set to True
        if 'B_toe' in column_names and 'Bm' in column_names and merge_Bm:
            to_export.Bm.fillna(to_export.B_toe, inplace=True)
            del to_export['B_toe']

        # merge slope_foundation with slope_toe if set to True
        if ('slope_toe' in column_names and 'slope_foundation' in column_names
                and merge_slope_toe):
            to_export.slope_toe.fillna(
                to_export.slope_foundation, inplace=True)
            del to_export['slope_foundation']

        # save to_export df to a excel file
        excel_save_name = f'{new_dir}/data.csv'
        to_export.to_csv(excel_save_name, index=False)

        print(f'folder {new_dir} is ready for Design Explorer 2')

    def get_concept(self, id=None, CaseNo=None):
        """ Get the specified concept

        Parameters
        ----------
        id : str, optional, default: None
            id of the concept, the id of a concept for a rubble mound
            breakwater out of rock is for instance: RRM.1
        CaseNo : int, optional, default: None
            CaseNo if the concept, only available if an export for the
            Design Explorer has been made with
            :py:meth:`to_design_explorer`

        Returns
        -------
        concept : obj
            a breakwater concept, for instance :py:obj:`RockRubbleMound`

        Raises
        ------
        InputError
            if a concept is not selected with CaseNo or id, or if there
            is no concept with the specified CaseNo or id
        KeyError
            if a concept can't be selected by CaseNo because the method
            :py:meth:`to_design_explorer` has not yet been used, and the
            CaseNo have therefore not yet been added to :py:attr:`df`
        """
        if id and CaseNo is not None:
            # cannot select concept by both id and CaseNo
            raise InputError(
                'Concept must be selected with id or CaseNo, not both')

        elif id is not None:
            # select concept by id
            type = id.split('.')[0]
            id = id.split('.')[1]

            row_ids = self.df[self.df['id'] == int(id)]
            row = row_ids[row_ids['type'] == type]

            msg = f'id {type}.{id}'

        elif CaseNo is not None:
            # check if CaseNo is a column in df,
            # is only added if the method to_design_explorer is used
            if 'CaseNo' not in self.df.columns:
                raise KeyError(
                    ('CaseNo will only be added to the df if an export for '
                     'the design explorer has been made, run '
                     'to_design_explorer or select by id'))

            # get the row of the df where the given CaseNo is in column CaseNo
            row = self.df[self.df.apply(
                lambda x: CaseNo in x['CaseNo'], axis=1)]

            msg = f'CaseNo {CaseNo}'

        else:
            # CaseNo and id is not given thus raise error
            raise InputError(
                ('No concept could be selected as neither an id nor CaseNo '
                 ' has been given'))

        # check if row has values
        if row.empty:
            # df is empty, thus CaseNo was invalid
            raise InputError(f'There is no concept with {msg}')
        else:
            # return the concept
            return row['concept'].values[0]

    def to_breakwaters(self, save_name):
        """ Save the df with designs in a pickle object

        Method will save the df with all breakwater concepts in a pickle
        object with extension .breakwaters.

        Parameters
        ----------
        save_name : str
            name of the file
        """
        # save to pickle
        file_name = f'{save_name}.breakwaters'

        file = open(file_name, 'wb')
        pickle.dump(self, file)

        print(f'saved to {file_name}')

    def show_warnings(self):
        """ Print all warnings encountered during the design

        Method prints a table containing the unique warning messages,
        with the time the warning was encountered during the design.
        """
        all_warnings = self.df.warnings

        unique_warnings = {}

        # iterate over de column in the df
        for warnings in all_warnings:
            # iterate over the list of warnings from each concept
            for warning in warnings:
                msg = str(warning.message)
                # check if warning is already encountered
                if msg in unique_warnings:
                    # not unique, thus add one to the count
                    unique_warnings[msg] += 1
                else:
                    # warning is unique, thus add to unique-wa
                    unique_warnings[msg] = 1

        # print the table
        table = [[key, count] for key, count in unique_warnings.items()]
        print(tabulate(table, headers=['Warning', 'Count'], tablefmt='github'))

    def cost_influence(self):
        """ Plot the influence of the varying parameters on the cost

        Method to see the influence of the varying parameters on the
        cost of a concept. If more than 1 parameter has been specified
        as a varying parameter the x-axis is normalised (from 0 to 1)
        so that all parameters can be plotted on the same axis.

        .. note::
           When more than one varying parameter is specified several
           lines must be plotted. To show the influence of one varying
           parameter on the cost the other varying parameters are set
           to their lower bound value.
        """
        # headers to exclude
        exclude = ['concept', 'id', 'type', 'warnings']

        # make dict to the line data for plotting
        lines = {}

        for structure in self.df.type.unique():
            # make df where the type equals the current structure
            df = self.df[self.df.type == structure]
            df = df.drop(columns=exclude)

            # iterate over the columns of the df
            for parameter in df.columns:
                # check if parameter is slope
                if 'slope' in parameter:
                    # check if not all nan
                    if not df[parameter].isnull().all():
                        # convert tuples to floats
                        df[parameter] = df[parameter].apply(
                            lambda x: np.round(np.arctan(x[0]/x[1])*180/np.pi, 2))

                # check if not a cost parameter or in exclude
                if parameter != 'cost' and parameter not in exclude:
                    # get the unique values
                    unique = df[parameter].unique()

                    # check if only 1 unique value in the column
                    if len(unique) > 1:
                        # varying parameter, add to lines dict
                        lines[parameter] = {'values': [], 'cost': []}

                        # get list of the values
                        values = list(df[parameter].values)

                        # iterate over the unique values
                        for unique_val in unique:
                            # get index of value in the df, first occurance
                            index = values.index(unique_val)

                            # get the row in the df with the data
                            row = df.iloc[index]

                            # add info to dict
                            # note that the unique value is retrieved from
                            # the df and not from the list with unique
                            # values, this is done so the result can be
                            # checked, i.e. if all values are indeed unique
                            lines[parameter]['values'].append(
                                row[parameter])

                            # compute average cost and add to dict
                            lines[parameter]['cost'].append(
                                np.mean(list(row.cost.values())))

        # generate the plot
        cost_influence(lines)



import re

from .exceptions import NotSupportedError

def _convert_string(string, mode, param):
    """ Convert a string to a list or tuple

    As a tuple and list is converted to a string by xlsxwriter it must
    be converted back to the desired type.

    Parameters
    ----------
    string : str
        the string to convert
    mode : {slope, lambda, varying}
        the type of the output
    param : str
        the parameter to convert
    """
    # validate input
    if not isinstance(string, str):
        raise TypeError(f'{param} is not correctly formatted in Excel')

    # get numbers from the string
    numbers = re.findall(r"[-+]?\d*\.\d+|\d+", string)

    # convert numbers to floats
    converted = []
    for number in numbers:
        converted.append(float(number))

    if mode == 'slope':
        # convert to tuple
        return tuple(converted)

    elif mode == 'lambda':
        return list(converted)

    elif mode == 'varying':
        pass

    else:
        raise NotSupportedError(
            f'Mode {mode} is not supported, must be slope, lambda or varying')



from ..utils.exceptions import user_warning, InputError, NotSupportedError

# supported armour layers, i.e. for these armour layers the rules for
# the underlayer and filter have been implemented
def _supported_armour_layers():
    """ returns the supported types of material for the armour layer """
    return ['Rock', 'Xbloc', 'XblocPlus']


def underlayer(Dn_armour, armour_layer, rho, rho_rock=2650):
    """ Design first underlayer of a rubble mound breakwater

    Design the layer of rock immediately below the armour layer, the
    nominal diameter of this layer (the underlayer) is determined using
    one of the following rules:

    - | Rock (CIRIA, CUR, CETMEF, 2007, p. 630)
      | :math:`M_{50u} = \\frac{M_{50a}}{15}` to :math:`\\frac{M_{50a}}{10}`

    - | Xbloc (Delta Marine Consultants, 2018)
      | :math:`M_{50u} = \\frac{M_{50a}}{15}` to :math:`\\frac{M_{50a}}{6}`

    - | XblocPlus (Delta Marine Consultants, 2018)
      | :math:`M_{50u} = \\frac{M_{50a}}{20}` to :math:`\\frac{M_{50a}}{8}`

    Parameters
    ----------
    Dn_armour : float
        nominal diameter of the armour layer [m]
    armour_layer : {'Rock', 'Xbloc', 'XblocPlus'}
        type of the armour layer
    rho : float, optional, default: 2650
        density of the armourstone [kg/m³]
    rho_rock : float, optional, default: 2650
        density of the rock used in the underlayer [kg/m³]

    Returns
    -------
    [float, float]
        list with the upper bound and lower bound limit of the Dn50 of
        the underlayer
    """
    M_armour = rho*Dn_armour**3

    if armour_layer == 'Rock':
        M_upper = M_armour/10
        M_lower = M_armour/15

    elif armour_layer == 'Xbloc':
        M_upper = M_armour/6
        M_lower = M_armour/15

    elif armour_layer == 'XblocPlus':
        M_upper = M_armour/8
        M_lower = M_armour/20

    else:
        supported = ', '.join(_supported_armour_layers())
        raise NotSupportedError(
            (f'{armour_layer} is not implemented as a rule for the underlayer'
             f', supported types are: {supported}.'))

    dn_upper = (M_upper/rho_rock)**(1/3)
    dn_lower = (M_lower/rho_rock)**(1/3)

    return [dn_lower, dn_upper]

def filter_layers(Dn, rho=2650):
    """ Design filter layer of a rubble mound breakwater

    Design a layer of rock below the first underlayer, or subsequent
    layers of rock. The nominal diameter of this layer (the filter layer)
    is given by the following rule:

    - | Rock (Van den Bos and Verhagen, 2018, p. 142)
      | :math:`M_{50l} = \\frac{M_{50u}}{25}` to :math:`\\frac{M_{50u}}{10}`

    Parameters
    ----------
    Dn : float
        nominal diameter of the layer above the filter layer [m]
    rho : float, optional, default: 2650
        density of the armourstone [kg/m³]

    Returns
    -------
    [float, float]
        list with the upper bound and lower bound limit of the Dn50 of
        the filter layer
    """
    M_upper_layer = Dn**3

    dn_upper = Dn/(10**(1/3))
    dn_lower = Dn/(25**(1/3))

    return [dn_lower, dn_upper]

def layer_coefficient(material, layers=None, placement=None):
    """ Get the layer thickness coefficient

    Determine the layer thickness coefficient, :math:`k_{t}`. In Table
    1 and 2 the included layer thickness coefficients can be found.

    Table 1: :math:`k_{t}` values for Rock (CIRIA, CUR, CETMEF, 2007, p. 126)

    +---------+-----------+-------+
    | Layers  | Placement | Value |
    +=========+===========+=======+
    | 1       | Dense     | 0.84  |
    +---------+-----------+-------+
    | 2       | Standard  | 0.91  |
    +---------+-----------+-------+
    | 2       | Dense     | 0.91  |
    +---------+-----------+-------+

    Table 2: :math:`k_{t}` values for Armour Units (CIRIA, CUR, CETMEF, 2007, p. 260)

    +-------------+--------+-------+
    | Armour Unit | Layers | Value |
    +=============+========+=======+
    | Cubes       |   2    |  1.1  |
    +-------------+--------+-------+
    | Tetrapods   |   2    |  1.02 |
    +-------------+--------+-------+
    | Dolos       |   2    |  0.94 |
    +-------------+--------+-------+
    | Accropode   |   1    |  1.29 |
    +-------------+--------+-------+
    | CoreLoc     |   1    |  1.52 |
    +-------------+--------+-------+
    | Xbloc       |   1    |  1.4  |
    +-------------+--------+-------+
    | XblocPlus   |   1    |  1.33 |
    +-------------+--------+-------+

    Parameters
    ----------
    material : str
        material of the layer
    layers : int, optional, default: None
        number of layers, required if the material is rock. For armour
        units the number of layers is not required, but a warning will
        be shown if the specified layer is different from the table
    placement : {Standard, Dense}, optional, default: None
        placement of the material, required if the material is rock

    Returns
    -------
    kt : float
        the layer thickness coefficient
    """
    # define dict for rock
    rock = {

        1: {'dense': 0.84},

        2: {'standard': 0.91, 'dense': 0.91}

    }

    # define dict for armour units
    units = {

        'Cubes': {'kt': 1.1, 'layers': 2},

        'Tetrapods': {'kt': 1.02, 'layers': 2},

        'Dolos': {'kt': 0.94, 'layers': 2},

        'Accropode': {'kt': 1.29, 'layers': 1},

        'CoreLoc': {'kt': 1.52, 'layers': 1},

        'Xbloc': {'kt': 1.4, 'layers': 1},

        'XblocPlus': {'kt': 1.33, 'layers': 1}

    }
    # Check if material has Accropode in the name,
    # as it can be Accropode I or Accropode II
    if 'Accropode' in material:
        # set material to Accropode
        material = 'Accropode'

    # get the kt value
    if material is 'Rock':
        # check if layer is given
        if layers is None:
            raise InputError(
                'Layers is a required argument if the material is Rock')

        # check if placement is given
        if placement is None:
            raise InputError(
                'Placement is a required argument if the material is Rock')

        # get the kt value
        kt = rock[layers][placement.lower()]

    elif material in list(units.keys()):
        # get kt value
        kt = units[material]['kt']

        # check layers
        num_layers = units[material]['layers']
        if layers is not None and layers != num_layers:
            # check number of layers
            if num_layers == 1:
                msg = (f'kt value for {material} is only determined for '
                       f'{num_layers} layer, not {layers}')
            else:
                # plural
                msg = (f'kt value for {material} is only determined for '
                       f'{num_layers} layers, not {layers}')

            user_warning(msg)

    else:
        # not supported armour layer
        kt = 1
        user_warning(f'{material} is not supported, continued with kt=1')

    return kt
